from __future__ import annotations

import tkinter as tk
import typing

from pygments.lexers._mapping import LEXERS

from biscuit.common import ActionSet
from biscuit.common.icons import Icons
from biscuit.common.textutils import *
from biscuit.common.ui import Frame

from .activitybar import ActivityBar
from .button import SButton

if typing.TYPE_CHECKING:
    from biscuit.editor import Text


class Statusbar(Frame):
    """Status bar
    Status bar is a container for status information and controls, displayed at the bottom of the application.
    Holds various widgets that are used to display information about the current file and editor state.
    """

    def __init__(self, master: Frame, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.layout.statusbar.background)

        self.activitybar = ActivityBar(self)
        self.activitybar.pack(side=tk.LEFT, padx=(10, 0))

        # self.terminal_toggle = self.add_button(
        #     icon="symbol-class",
        #     callback=self.toggle_sidebar,
        #     description="Toggle terminal",
        #     side=tk.LEFT,
        # )
        # self.terminal_toggle.config(padx=10)
        # self.terminal_toggle.show()

        # TODO: making this more , Make default lists of buttons, map them with pack(bool), callback
        # also make an active_editor-specific/global lists of buttons to show/hide in bulk

        # ---------------------------------------------------------------------
        self.branch = self.add_button(
            text="main",
            icon=Icons.SOURCE_CONTROL,
            callback=self.base.commands.change_git_branch,
            description="Checkout branch",
            side=tk.LEFT,
            padx=(2, 0),
        )

        # ---------------------------------------------------------------------
        self.process_indicator = self.add_button(
            text="setting up environment",
            icon=Icons.SYNC,
            description="enabling language extensions",
            side=tk.LEFT,
            padx=(2, 0),
        )

        # ---------------------------------------------------------------------
        self.lc_actionset = self.create_actionset(
            "Goto line in active editor",
            ":",
            pinned=[["goto line: {}", self.goto_line]],
        )
        self.line_col_info = self.add_button(
            text="1,1",
            callback=self.base.commands.goto_line_column,
            description="Go to Line/Column",
            side=tk.RIGHT,
        )

        # ---------------------------------------------------------------------
        self.indent_actionset = self.create_actionset(
            "Change indentation",
            "indent:",
            [
                ("2 spaces", lambda e=None: self.base.set_tab_spaces(2)),
                ("4 spaces", lambda e=None: self.base.set_tab_spaces(4)),
                ("8 spaces", lambda e=None: self.base.set_tab_spaces(8)),
            ],
            pinned=[
                ["custom: {} spaces", self.change_custom_indentation],
            ],
        )
        self.indentation = self.add_button(
            text=f"{self.base.tab_spaces}sp",
            callback=self.base.commands.change_indentation_level,
            description="Change indentation",
            side=tk.RIGHT,
        )

        # ---------------------------------------------------------------------
        self.encoding_actionset = self.create_actionset(
            "Change file encoding",
            "encoding:",
            [("UTF-8", lambda e=None: print("encoding UTF-8", e))],
        )
        self.encoding = self.add_button(
            text="UTF-8",
            callback=self.base.commands.change_encoding,
            description="Change encoding",
            side=tk.RIGHT,
        )

        # ---------------------------------------------------------------------
        self.eol_actionset = self.create_actionset(
            "Change End of Line sequence",
            "eol:",
            [(i.upper(), self.change_eol(val)) for i, val in eol_map.items()],
        )
        self.eol = self.add_button(
            text="CRLF",
            callback=self.base.commands.change_end_of_line_character,
            description="Change End of Line sequence",
            side=tk.RIGHT,
        )

        # ---------------------------------------------------------------------
        items = [
            (aliases[0], self.change_language(aliases[0]))
            for _, _, aliases, _, _ in LEXERS.values()
            if aliases
        ]
        self.language_actionset = self.create_actionset(
            "Change Language Mode", "language:", items
        )
        self.file_type = self.add_button(
            text="Plain Text",
            callback=self.base.commands.change_language_mode,
            description="Change Language Mode",
            side=tk.RIGHT,
        )

        # ---------------------------------------------------------------------

        self.secondary_activitybar = ActivityBar(self)
        self.secondary_activitybar.pack(side=tk.RIGHT, padx=(0, 10))

        # ---------------------------------------------------------------------

        self.panel_toggle = SButton(
            self,
            icon=Icons.LAYOUT_PANEL_OFF,
            callback=self.toggle_panel,
            description="Toggle panel",
            icon2=Icons.LAYOUT_PANEL,
        )
        self.panel_toggle.set_pack_data(side=tk.RIGHT, padx=(0, 10))

        self.panel_toggle.show()

    def add_button(
        self,
        text="",
        icon="",
        callback: typing.Callable = None,
        description="",
        highlighted=False,
        **kwargs,
    ) -> SButton:
        """Creates and configures an SButton.

        Args:
            text (str): The text to display on the button.
            icon (str, optional): The icon to display on the button.
            callback (typing.Callable, optional): The callback to call when the button is clicked.
            description (str, optional): The description of the button to display on hover.
            side (str, optional): The side to pack the button on. Defaults to tk.LEFT.
            padx (tuple, optional): The padding for the button on the x-axis. Defaults to (2, 0).

        Returns:
            SButton: The created SButton instance.
        """

        btn = SButton(
            self,
            text=text,
            icon=icon,
            callback=callback,
            description=description,
            highlighted=highlighted,
        )
        btn.set_pack_data(**kwargs)
        return btn

    def create_actionset(
        self, name: str, prefix: str, actions: list = [], pinned: list = []
    ) -> ActionSet:
        """Creates and registers an ActionSet.

        Args:
            name (str): The name of the action set.
            prefix (str): The prefix for the action set.
            actions (list): A list of actions to include in the action set.

        Returns:
            ActionSet: The created ActionSet instance.
        """
        actionset = ActionSet(name, prefix, actions, pinned=pinned)
        self.base.palette.register_actionset(lambda: actionset)
        return actionset

    def toggle_sidebar(self) -> None:
        """Toggles the sidebar visibility."""

        self.base.root.toggle_sidebar()

    def toggle_panel(self) -> None:
        """Toggles the panel visibility."""

        self.base.toggle_terminal()

    def toggle_editmode(self, state: bool) -> None:
        """Toggles the edit mode for the status bar.

        Args:
            state (bool): If True, shows edit mode related widgets. Otherwise, hides them.
        """

        widgets = [
            self.file_type,
            # TODO: EOL, Encoding should be optional
            # self.eol,
            # self.encoding,
            self.indentation,
            self.line_col_info,
        ]
        for widget in widgets:
            widget.show() if state else widget.hide()

    def update_git_info(self) -> None:
        """Updates the Git branch information displayed on the status bar."""

        if self.base.git_found:
            self.branch.show()
            self.branch.change_text(f"{self.base.git.active_branch}")
        else:
            self.branch.hide()

    def on_open_file(self, text: Text) -> None:
        """Updates the status bar information when a file is opened.

        Args:
            text (Text): The text object representing the opened file.
        """

        self.file_type.change_text(text.language)
        # self.encoding.change_text(text.encoding.upper())
        # self.eol.change_text(get_eol_label(text.eol))

    def set_line_col_info(self, line: int, col: int, selected: int = None) -> None:
        """Sets the line and column information on the status bar.

        Args:
            line (int): The current line number.
            col (int): The current column number.
            selected (int): The number of selected characters.
        """

        self.line_col_info.change_text(
            f"{line},{col} "  # + (f"({selected})" if selected else "")
        )

    def set_encoding(self, encoding: str) -> None:
        """Sets the file encoding displayed on the status bar.

        Args:
            encoding (str): The encoding to display.
        """

        self.encoding.change_text(text=encoding.upper())

    def set_spaces(self, spaces: int) -> None:
        """Sets the indentation spaces displayed on the status bar.

        Args:
            spaces (int): The number of spaces to set for indentation.
        """

        self.indentation.change_text(text=f"{spaces}sp")

    def pack(self):
        """Packs the status bar into the application."""

        super().pack(fill=tk.X, pady=(1, 0))

    def goto_line(self, line: str) -> None:
        """Goes to a specific line in the active editor.

        Args:
            line (str): The line number to go to.
        """

        if line and line.isnumeric():
            self.base.editorsmanager.active_editor.content.goto_line(int(line))
        else:
            print("failed goto line", line)

    def change_custom_indentation(self, line: str = None) -> None:
        """Changes the indentation to a custom number of spaces.

        Args:
            line (str, optional): The number of spaces for indentation.
        """

        if line and line.strip().isnumeric():
            self.base.set_tab_spaces(int(line.strip()))
        else:
            print("failed change indentation", line)

    def change_eol(self, val: str) -> typing.Callable:
        """Palette helper function
        Changes the end-of-line sequence.

        Args:
            val (str): The end-of-line sequence to change to.

        Returns:
            typing.Callable: A callable that changes the end-of-line sequence.
        """

        return lambda _: self.base.editorsmanager.active_editor.content.text.change_eol(
            eol=val
        )

    def change_language(self, language: str) -> typing.Callable:
        """Palette helper function
        Changes the language mode of the active editor.

        Args:
            language (str): The language mode to change to.

        Returns:
            typing.Callable: A callable that changes the language mode.
        """

        return lambda _: self.base.editorsmanager.active_editor.content.text.highlighter.change_language(
            language
        )



from .endpoint import Endpoint


class Theme(Endpoint):
    def __init__(self, *a) -> None:
        super().__init__(*a)

    # TODO themes registration, api
    def register_theme(self, theme): ...

    def set_theme(self, theme): ...



from __future__ import annotations

import os
import threading
import time
import typing
from importlib import util
from pathlib import Path
from queue import Queue

import requests
import toml

from biscuit.extensions.viewer.viewer import ExtensionViewer
from biscuit.views.extensions.extension import ExtensionGUI

if typing.TYPE_CHECKING:
    from biscuit import App

    from . import extension


class ExtensionManager:
    """Extension manager for Biscuit.

    Manages the loading and execution of extensions.
    """

    def __init__(self, base: App) -> None:
        self.base: App = base

        self.extensions_loaded = False

        self.interval = 5
        self.installed: dict[str, bool] = {}

        self.repo_url = (
            "https://raw.githubusercontent.com/tomlin7/biscuit-extensions/main/"
        )
        self.list_url = self.repo_url + "extensions.toml"

        self.fetched: dict[str, str] = {}
        self.fetch_queue = Queue()
        self.fetching = threading.Event()
        self.extensions_lock = threading.Lock()

        if not (self.base.extensiondir and os.path.isdir(self.base.extensiondir)):
            print("Extensions directory not found.")
            return

        self.load_queue = Queue()

    def install_extension_from_name(self, name: str) -> bool:
        """Install an extension from the repository by name."""

        if name in self.fetched:
            data = self.fetched[name]

            t = self.run_fetch_extension(
                ExtensionGUI(self.base.extensions_view.results, name, data)
            )
            t.join()

            return True

        return False

    def uninstall_extension_from_name(self, name: str) -> bool:
        """Uninstall an extension by name."""

        if name in self.installed:
            self.remove_extension(self.installed[name])
            return True

        return False

    def find_extension_by_name(self, name: str) -> str:
        """Search for an extension by name."""

        return self.fetched.get(name, None)

    def list_all_extensions(self) -> typing.Generator[tuple[str, str]]:
        """List all extensions."""

        for name, data in self.fetched.items():
            yield name, data

    def list_installed_extensions(self) -> typing.Generator[tuple[str, str]]:
        """List installed extensions."""

        for name in self.installed:
            yield name, self.fetched.get(name, None)

    def list_extensions_by_user(self, user: str) -> typing.Generator[tuple[str, str]]:
        """Show extensions by a specific user."""

        for id, data in self.fetched.items():
            if data[1] == user:
                yield id, data

    def run_fetch_extensions(self, *_) -> None:
        """Called from the Extensions View to fetch extensions."""

        if self.base.testing:
            return

        if self.fetching.is_set():
            self.fetching.wait()

        with self.extensions_lock:
            threading.Thread(target=self.fetch_extensions, daemon=True).start()

    def fetch_searched_extensions(self, search_string) -> None:
        """Fetch extensions whose name matches the search_string from Extensions repository."""

        response = None
        try:
            response = requests.get(self.list_url)
        except Exception:
            pass

        # FAIL - network error
        if not response or response.status_code != 200:
            try:
                self.base.extensions_view.results.show_placeholder()
            except:
                ...
            return

        self.fetched = toml.loads(response.text)
        # SUCCESS
        if self.fetched:
            try:
                self.base.extensions_view.results.show_content()
            except:
                ...

        if not search_string:
            for name, data in self.fetched.items():
                self.fetch_queue.put((name, data))
        else:
            for name, data in self.fetched.items():
                if search_string.lower() in name.lower():
                    self.fetch_queue.put((name, data))

    def fetch_extensions(self) -> None:
        """Fetch extensions from Extensions repository."""

        response = None
        try:
            response = requests.get(self.list_url)
        except Exception:
            pass

        # FAIL - network error
        if not response or response.status_code != 200:
            try:
                self.base.extensions_view.results.show_placeholder()
            except:
                ...
            return

        self.fetched = toml.loads(response.text)
        # SUCCESS
        if self.fetched:
            try:
                self.base.extensions_view.results.show_content()
            except:
                ...

        for name, data in self.fetched.items():
            # TODO add further loops for folders
            self.fetch_queue.put((name, data))

    def run_fetch_extension(self, ext: ExtensionGUI) -> None:
        """Called from the Extension View to fetch an extension."""

        if ext.installed:
            return

        ext.set_fetching()

        t = threading.Thread(target=self.fetch_extension, args=(ext,), daemon=True)
        t.start()
        return t

    def fetch_extension(self, ext: ExtensionGUI) -> None:
        """Fetch an extension from the repository."""

        try:
            response = requests.get(ext.url)
            if response.status_code == 200:
                self.save_extension(ext, response)
        except:
            ext.set_unavailable()

    def save_extension(self, ext: ExtensionGUI, response: requests.Response) -> None:
        """Save a fetched extension. Saves the extension and loads it."""

        with open(ext.file, "w") as fp:
            fp.write(response.text)

        self.load_extension(ext.file)
        ext.set_installed()

        self.base.logger.info(f"Fetching extension '{ext.name}' successful.")
        self.base.notifications.info(f"Extension '{ext.name}' has been installed!")

    def remove_extension(self, ext: ExtensionGUI) -> None:
        """Remove an extension from the system."""

        try:
            os.remove(ext.file)
            self.unload_extension(ext.filename)

            ext.set_uninstalled()

            self.base.logger.info(f"Uninstalling extension '{ext.name}' successful.")
            self.base.notifications.info(
                f"Extension '{ext.name}' has been uninstalled!"
            )
        except Exception as e:
            self.base.logger.error(f"Uninstalling extension '{ext.name}' failed.\n{e}")

    def open_extension(self, ext: ExtensionGUI) -> None:
        """Open extension details."""

        self.base.extension_viewer.show(ext)

    def load_extension(self, file: str):
        """Load an extension from a file."""

        if file.endswith(".py"):
            extension_name = os.path.basename(file).split(".")[0]

            # if an extension with same name is already loaded, skip
            if extension_name in self.installed:
                return
            self.load_queue.put((extension_name, file))

    def unload_extension(self, file: str):
        """Unload an extension from a file."""

        if file.endswith(".py"):
            file = os.path.splitext(file)[0]

        if file in self.installed:
            self.installed.pop(file)

    def load_extensions(self):
        """Not to be called directly. Use `start_server` instead."""

        refresh_count = 0
        while self.alive:
            if self.load_queue.empty():
                self.extensions_loaded = True
                time.sleep(self.interval)

                refresh_count += 1
                if refresh_count == 60:
                    self.base.logger.info(
                        f"Extensions server: active, {len(self.installed)} extensions loaded."
                    )
                    refresh_count = 0
                continue

            ext, path = self.load_queue.get()
            module_name = f"extensions.{ext}"
            try:
                # TODO support for multiple files
                spec = util.spec_from_file_location(module_name, str(path))
                extension_module: extension = util.module_from_spec(spec)

                spec.loader.exec_module(extension_module)
                # execute the setup function
                extension_module.setup(self.base.api)

                self.base.logger.info(f"Extension '{ext}' loaded.")
            except Exception as e:
                self.base.logger.error(f"Failed to load extension '{ext}': {e}")
                self.base.notifications.error(f"Extension '{ext}' failed: see logs.")
                if ext in self.installed:
                    self.installed.pop(ext)

    def register_installed(self, name: str, extension: object) -> None:
        """Register an installed extension."""

        self.installed[name] = extension
        try:
            extension.install()
        except Exception as e:
            print(e)
            # most likely the extension does not implement the install method
            ...

    def queue_installed_extensions(self) -> None:
        def load_extension(dir: str):
            if not os.path.isdir(dir):
                return

            for extension_file in os.listdir(dir):
                if extension_file.endswith(".py"):
                    extension_name = os.path.basename(extension_file).split(".")[0]
                    if extension_name in self.installed:
                        return

                    self.load_queue.put((extension_name, Path(dir) / extension_file))

        load_extension(self.base.extensiondir)
        load_extension(self.base.fallback_extensiondir)

    def start_server(self):
        self.queue_installed_extensions()
        self.alive = True
        self.server = threading.Thread(target=self.load_extensions, daemon=True)
        self.server.start()

        self.base.logger.info(f"Extensions server started.")

    def restart_server(self):
        self.installed.clear()
        self.queue_installed_extensions()

    def stop_server(self):
        print(f"Extensions server stopped.")
        self.alive = False

    # def restricted_import(self, name, globals={}, locals={}, fromlist=[], level=0):
    #     # sandboxed import function
    #     self.blocked_modules = ['os', 'sys']
    #     self.imports = {module: __import__(module) for module in sys.modules}
    #     sys.modules['builtins'].__import__ = self.restricted_import

    #     if name in self.blocked_modules:
    #         raise ImportError("Module '{}' is not allowed.".format(name))

    #     return self.imports[name]



from __future__ import annotations

import typing

import click

from biscuit.git import URL

if typing.TYPE_CHECKING:
    from biscuit import App


@click.command()
@click.argument("url", type=str)
def clone(url) -> typing.Callable[[App, str], None]:
    """Clone & open a git repository in Biscuit

    This command will clone a git repository and open it in a new window.

    Example:
        biscuit clone

    Args:
        url (str): The url of the git repository
    """

    if not url:
        url = click.prompt("Git repository url", type=str)

    click.echo(
        f"Cloning repository from {'https://github.com/' if not URL.match(url) else ''}{url}"
    )
    return lambda app, url=url: app.clone_repo(url, new_window=False)


@click.command()
@click.argument(
    "file1",
    type=click.Path(
        exists=True,
        dir_okay=False,
        resolve_path=True,
    ),
    required=False,
)
@click.argument(
    "file2",
    type=click.Path(
        exists=True,
        dir_okay=False,
        resolve_path=True,
    ),
    required=False,
)
def diff(file1=None, file2=None) -> typing.Callable[[App, str], None]:
    """Diff two files

    This command will open a new window with the diff of the two files.

    Example:
        biscuit diff path/to/file path/to/second/file

    Args:
        file1 (str, optional): The path to the first file. Defaults to None.
        file2 (str, optional): The path to the second file. Defaults to None."""

    if not file1:
        file1 = click.prompt("path/to/file", type=str)
    if not file2:
        file2 = click.prompt("path/to/second/file", type=str)

    return lambda app, file1=file1, file2=file2: app.diff_files(file1, file2)


def register(cli: click.Group):
    cli.add_command(clone)
    cli.add_command(diff)



import tkinter as tk

from biscuit.common.ui import Closable, Frame
from biscuit.common.ui.native import Label, Toplevel


class OpenEditors(Toplevel):
    """Toplevel view that displays the open editors in the editor tabs.

    The OpenEditors view displays the open editors in the editor tabs.
    - The user can switch between open editors, close the editors."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)
        self.overrideredirect(True)
        self.nodes = {}
        self.withdraw()

        self.container = Frame(self)
        self.container.pack(fill=tk.BOTH, expand=True, padx=1, pady=1)

        Label(
            self.container,
            text="Opened Editors",
            font=self.base.settings.uifont,
            padx=10,
            pady=5,
            **self.base.theme.utils.label,
        ).pack(fill=tk.X, expand=True, anchor=tk.W)

        self.bind("<FocusOut>", lambda _: self.hide())

    def show(self, e: tk.Event) -> None:
        btn = e.widget
        x, y = (
            btn.winfo_rootx() - self.winfo_width() + btn.winfo_width(),
            btn.winfo_rooty() + btn.winfo_height(),
        )
        self.geometry(f"+{x}+{y}")
        self.deiconify()
        self.update()
        self.focus_force()

    def hide(self) -> None:
        self.withdraw()

    def add_item(self, editor):
        temp = Closable(
            self.container,
            text=editor.filename or editor.path,
            callback=lambda p=editor.path: self.openfile(p),
            closefn=lambda p=editor.path: self.closefile(p),
        )
        try:
            temp.text_label.config(anchor=tk.W)
        except AttributeError:
            pass

        temp.pack(fill=tk.X, expand=True)
        self.nodes[editor.path] = temp
        self.update()

    def remove_item(self, editor):
        if not self.nodes:
            return

        e = self.nodes.pop(editor.path)
        e.pack_forget()
        e.destroy()
        self.update()

    def set_active(self, editor):
        # TODO: set highlight, clear highlight on others
        ...

    def clear(self):
        for node in self.nodes.values():
            node.destroy()
        self.nodes = {}

    def openfile(self, path) -> None:
        self.base.editorsmanager.editorsbar.switch_tabs(path)

    def closefile(self, path) -> None:
        e = self.base.editorsmanager.close_editor_by_path(path)
        self.base.editorsmanager.editorsbar.close_tab_helper(e)



from __future__ import annotations

import tkinter as tk
from typing import Callable, Union

from biscuit.common.menu import submenu
from biscuit.common.menu.submenu import SubMenuItem

from ..ui import Frame, Toplevel
from .checkable import Checkable
from .command import Command
from .separator import Separator

# TODO Menus
# - Have various types of menus
# - Context menus for various buttons across the editor
# - Make the api more clean and easy to use like palette
#   - isolate tkinter_menus library


class Menu(Toplevel):
    """Base class for all menus
    Underlying is a Toplevel widget with a container frame
    that holds the menu items and separators.

    Supports adding commands, checkables and separators"""

    submenu = False

    def __init__(self, master, name: str = None, *args, **kwargs) -> None:
        """Create a new menu

        Args:
            master (tk.Tk): The master widget
            name (str, optional): The name of the menu. Defaults to None.
        """

        super().__init__(master, *args, **kwargs)
        self.active = False
        self.name = name

        self.config(bg=self.base.theme.border)
        self.withdraw()
        self.overrideredirect(True)

        self.container = Frame(self, padx=5, pady=5, **self.base.theme.menu)
        self.container.pack(fill=tk.BOTH, expand=True, padx=1, pady=1)
        self.container.grid_columnconfigure(0, weight=1)
        self.container.grid_rowconfigure(0, weight=1)
        self.container.hide = self.hide

        self.menu_items = []
        self.row = 0

        self.hold_focus = False

        self.bind("<FocusOut>", self.hide_check)
        self.bind("<Escape>", self.hide)

    def get_coords(self, *e) -> tuple:
        """This method is to be overridden by subclasses to provide
        custom placement of the menu that may or may not depend on the event.
        The method may take the event as arguments and must return
        the x, y coordinates as a tuple of integers.

        Args:
            *e: The event that triggered the menu

        Returns:
            tuple: The x, y coordinates
        """

        return (
            self.master.winfo_rootx(),
            self.master.winfo_rooty() + self.master.winfo_height(),
        )

    def show(self, *e) -> None:
        """Show the menu at the given event coordinates
        For relative placement, override the get_coords method
        in the subclass and also bind the event to this method.

        Args:
            *e: The event that triggered the menu"""

        self.active = True
        self.update_idletasks()

        x, y = self.get_coords(*e)
        self.wm_geometry(f"+{x}+{y}")

        self.deiconify()
        self.focus_set()

    def hide(self, *args) -> None:
        """Hide the menu and set the active flag to False"""

        self.active = False
        self.withdraw()
        self.master.event_generate("<<Hide>>")

    def hide_check(self, e) -> None:
        """If submenus are active, dont hide the menu"""

        if self.hold_focus:
            return

        self.hide()

    def add_item(self, item: Union[Command, Checkable]) -> Command:
        """Add a menu item to the menu
        Not to be confused with the add_command or add_checkable methods.
        This method is used to add an item that is already created to the menu.

        Args:
            item (Union[Command, Checkable]): The menu item to add
        """

        item.grid(row=self.row, sticky=tk.EW, pady=0)
        self.menu_items.append(item)

        self.row += 1
        return item

    def add_checkable(self, text, command=lambda *_: ..., checked=False) -> Checkable:
        """Add a checkable item to the menu

        Args:
            text (str): The text to display on the menu item
            command (Callable, optional): The command to run when the item is clicked. Defaults to lambda *_:....
            checked (bool, optional): The initial checked state. Defaults to False.

        Returns:
            Checkable: The created checkable menu item"""

        new_item = Checkable(self.container, text, command, checked=checked)
        return self.add_item(new_item)

    def add_command(
        self, text: str, command: Callable = lambda *_: ..., *args, **kwargs
    ) -> Command:
        """Add a command to the menu

        Args:
            text (str): The text to display on the menu item
            command (Callable, optional): The command to run when the item is clicked. Defaults to lambda *_:....

        Returns:
            Command: The created menu item"""

        new_item = Command(self.container, text, command, *args, **kwargs)
        return self.add_item(new_item)

    def add_separator(self, length=18) -> Separator:
        """Add a separator to the menu

        Args:
            length (int, optional): The length of the separator. Defaults to 18.

        Returns:
            Separator: The created separator"""

        new_sep = Separator(self.container, length)
        new_sep.grid(row=self.row, sticky=tk.EW, pady=0)
        self.menu_items.append(new_sep)

        self.row += 1
        return new_sep

    def add_menu(self, text: str) -> SubMenuItem:
        """Add a new menu to the menu

        Args:
            text (str): The text of the menu

        Returns:
            SubMenuItem: The created menu item"""

        new_item = SubMenuItem(self, self.container, text)
        return self.add_item(new_item)

    def clear(self) -> None:
        """Clear all menu items from the menu"""

        for item in self.menu_items:
            item.destroy()

        self.menu_items = []
        self.row = 0

    def event_chosen(self, *_) -> None:
        """Event handler for when a menu item is chosen"""

        self.hide()



from __future__ import annotations

import typing

from biscuit.common import Menu
from biscuit.common.ui import Menubutton

if typing.TYPE_CHECKING:
    from . import Menubar


class MenubarItem(Menubutton):
    def __init__(self, master: Menubar, text, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.name = text
        self.config(
            text=text,
            padx=10,
            pady=5,
            font=self.base.settings.uifont,
            **self.base.theme.layout.menubar.item,
        )

        self.menu = Menu(self, self.name)
        self.bind("<<Hide>>", self.deselect)
        self.bind("<Button-1>", self.onclick)
        self.bind("<Enter>", self.hover)

    def onclick(self, *_):
        self.menu.show()
        self.select()

    def hover(self, *_):
        self.master.switch_menu(self)

    def select(self):
        self.config(bg=self.base.theme.layout.menubar.item.highlightbackground)

    def deselect(self, *_):
        self.config(bg=self.base.theme.layout.menubar.item.background)



from biscuit.common.icons import Icons

kinds = (
    (Icons.SYMBOL_KEY, None),  # text
    (Icons.SYMBOL_METHOD, "#AB7CCF"),  # method
    (Icons.SYMBOL_METHOD, "#AB7CCF"),  # function
    (Icons.SYMBOL_METHOD, "#6eb1ee"),  # constructor
    (Icons.SYMBOL_FIELD, "#6eb1ee"),  # field
    (Icons.SYMBOL_VARIABLE, "#6eb1ee"),  # variable
    (Icons.SYMBOL_CLASS, "#de9327"),  # class
    (Icons.SYMBOL_INTERFACE, "#6eb1ee"),  # interface
    (Icons.SYMBOL_NAMESPACE, None),  # module
    (Icons.SYMBOL_PROPERTY, "#6eb1ee"),  # property
    (Icons.SYMBOL_RULER, None),  # unit
    (Icons.SYMBOL_NUMERIC, None),  # value
    (Icons.SYMBOL_ENUM, None),  # enum
    (Icons.SYMBOL_KEYWORD, None),  # keyword
    (Icons.SYMBOL_SNIPPET, None),  # snippet
    (Icons.SYMBOL_COLOR, "#ffd700"),  # color
    (Icons.SYMBOL_FILE, None),  # file
    (Icons.SYMBOL_CLASS, None),  # reference
    (Icons.FOLDER, None),  # folder
    (Icons.SYMBOL_ENUM_MEMBER, None),  # enummember
    (Icons.SYMBOL_CONSTANT, None),  # constant
    (Icons.SYMBOL_STRUCTURE, None),  # struct
    (Icons.SYMBOL_EVENT, "#ffd700"),  # event
    (Icons.SYMBOL_OPERATOR, None),  # operator
    (Icons.SYMBOL_PARAMETER, "#ffd700"),  # typeparameter
)



from __future__ import annotations

import os
import sqlite3
import tkinter as tk
import typing

from biscuit.common import Dropdown
from biscuit.common.chat import ChatModelInterface, Gemini1p5Flash
from biscuit.common.icons import Icons

from ..sidebar_view import SideBarView
from .chat import Chat
from .menu import AIMenu
from .placeholder import AIPlaceholder

if typing.TYPE_CHECKING:
    ...


PROMPT = """You are an assistant part of Biscuit code editor named Bikkis. 
You are an expert programmer, you'll help the user with his queries.

Biscuit have many features like auto completions, git integration,
code diagnostics, code refactoring, code navigation, advanced search, etc.
It supports many languages like Python, JavaScript, Java, C++, etc.
User can install extensions right within the editor's extension marketplace.
Don't make assumptions about the existence of a feature in Biscuit.

Give the user minimal responses and be straightforward. 
If there are files attached by user, they will be appended to the question.
You will have to consider them while responding.

Reply to this message from user: """


class AI(SideBarView):
    """A view that displays the AI chat.

    The AI view allows the user to chat with an AI using the Gemini API.
    - The user can configure the API key to use for the chat, which is stored in the secrets database.
    - The chat can be refreshed to start a new chat."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.__icon__ = Icons.SPARKLE_FILLED
        self.name = "AI"
        self.chat = None
        self.api_key = ""

        self.title.grid_forget()

        self.api_providers = {
            "Gemini 1.5 Flash": Gemini1p5Flash,
        }
        self.current_provider = "Gemini 1.5 Flash"

        self.dropdown = Dropdown(
            self.top,
            items=self.api_providers.keys(),
            selected=self.current_provider,
            callback=self.set_current_provider,
        )
        self.top.grid_columnconfigure(self.column, weight=1)

        self.dropdown.grid(row=0, column=0, sticky=tk.NSEW, padx=(0, 10))

        self.menu = AIMenu(self)
        self.menu.add_command("New Chat", self.new_chat)
        self.menu.add_command("Configure API Key...", self.add_placeholder)

        self.add_action(Icons.REFRESH, self.new_chat)
        self.add_action(Icons.ELLIPSIS, self.menu.show)

        self.db = sqlite3.connect(os.path.join(self.base.datadir, "secrets.db"))
        self.cursor = self.db.cursor()
        self.cursor.executescript(
            """
            CREATE TABLE IF NOT EXISTS secrets (key TEXT PRIMARY KEY NOT NULL, value TEXT);
            """
        )

        self.cursor.execute("SELECT value FROM secrets WHERE key='GEMINI_API_KEY'")
        api_key = self.cursor.fetchone()

        self.placeholder = AIPlaceholder(self)
        if api_key:
            self.add_chat(api_key[0])
        else:
            self.add_placeholder()

    def register_provider(self, provider: str, model: ChatModelInterface) -> None:
        """Register a new provider for the chat.

        Args:
            provider (str): The provider to register."""

        self.api_providers[provider] = model
        self.dropdown.add_command(provider)

    def set_current_provider(self, provider: str) -> None:
        """Set the current provider for the chat.

        Args:
            provider (str): The provider to set as the current provider."""

        if provider == self.current_provider:
            return

        self.current_provider = provider
        self.add_chat()

    def attach_file(self, *files: typing.List[str]) -> None:
        """Attach a file to the chat.

        Args:
            files (list): The list of files to attach to the chat."""

        if self.chat:
            self.chat.attach_file(*files)

    def add_placeholder(self) -> None:
        """Show the home page for the AI assistant view"""

        self.add_item(self.placeholder)
        if self.api_key:
            self.placeholder.api_key.set(self.api_key)

        if self.chat:
            self.remove_item(self.chat)
            self.chat.destroy()

    def add_chat(self, api_key: str = None) -> None:
        """Add a new chat to the view.

        Args:
            api_key (str): The API key to use for the chat. Defaults to configured."""

        if api_key:
            self.api_key = api_key

        if not self.api_key:
            return self.add_placeholder()

        self.cursor.execute(
            "INSERT OR REPLACE INTO secrets (key, value) VALUES ('GEMINI_API_KEY', ?)",
            (self.api_key,),
        )
        self.db.commit()

        if self.chat:
            self.remove_item(self.chat)
            self.chat.destroy()
            self.chat = None

        self.chat = Chat(self)
        self.chat.set_model(self.get_model_instance(PROMPT))
        self.add_item(self.chat)
        self.remove_item(self.placeholder)

    def get_model_instance(self, prompt: str) -> ChatModelInterface:
        """Get the model instance for the current provider.

        Returns:
            ChatModelInterface: The model instance for the current provider."""

        return self.api_providers[self.current_provider](self.api_key, prompt)

    def new_chat(self) -> None:
        """Start a new chat with the AI assistant."""

        if self.chat:
            try:
                return self.chat.new_chat()
            except Exception:
                pass

        self.add_chat()



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from . import App


class Binder:
    """Class for binding events to keys"""

    def __init__(self, base: App) -> None:
        self.base = base

        self.bindings = self.base.settings.bindings
        self.events = self.base.commands

        self.bind_all()

    def bind_all(self) -> None:
        """Bind all bindings from config"""

        self.bind(self.bindings.new_file, self.events.new_file)
        self.bind(self.bindings.new_window, self.events.new_window)
        self.bind(self.bindings.open_file, self.events.open_file)
        self.bind(self.bindings.open_dir, self.events.open_directory)
        self.bind(self.bindings.save, self.events.save_file)
        self.bind(self.bindings.save_as, self.events.save_file_as)
        self.bind(self.bindings.close_file, self.events.close_editor)
        self.bind(self.bindings.quit, self.events.quit_biscuit)
        self.bind(self.bindings.undo, self.events.undo)
        self.bind(self.bindings.redo, self.events.redo)
        self.bind(
            self.bindings.restore_closed_tab, self.events.restore_last_closed_editor
        )
        self.bind(self.bindings.close_all_tabs, self.events.close_all_editors)
        self.bind(self.bindings.change_tab, self.events.change_tab_forward)
        self.bind(self.bindings.change_tab_back, self.events.change_tab_back)
        self.bind(self.bindings.split_tab, self.events.split_editor)

    def late_bind_all(self) -> None:
        """Bindings that require full initialization"""

        self.bind(self.bindings.command_palette, self.events.show_command_palette)
        self.bind(self.bindings.file_search, self.events.search_files)
        self.bind(self.bindings.symbol_outline, self.events.show_symbol_palette)
        self.bind(self.bindings.goto_line, self.events.goto_line_column)

        self.bind(self.bindings.panel, self.base.contentpane.toggle_panel)
        self.bind(self.bindings.sidebar, self.events.toggle_sidebar)
        self.bind(self.bindings.secondary_sidebar, self.events.toggle_secondary_sidebar)
        self.bind(self.bindings.directory_tree, self.events.show_directory_tree)
        self.bind(self.bindings.extensions, self.events.show_extensions)
        self.bind(self.bindings.global_search, self.events.show_search)
        self.bind(self.bindings.debugger, self.events.show_debugger)
        self.bind(self.bindings.git, self.events.show_source_control)
        self.bind(self.bindings.assistant, self.events.show_assistant)
        self.bind(self.bindings.logs, self.events.show_logs)

        self.bind(self.bindings.open_settings, self.events.open_settings)
        self.bind(
            self.bindings.restore_recent_session, self.events.restore_recent_session
        )
        self.bind(self.bindings.open_recent_folders, self.events.show_recent_folders)
        self.bind(self.bindings.open_recent_files, self.events.show_recent_files)
        self.bind(
            self.bindings.open_recent_session, self.events.restore_last_closed_editor
        )

    def bind(self, this, to_this) -> None:
        self.base.bind(this, to_this)



import queue
import tkinter as tk
from datetime import datetime

from biscuit.common import caller_class_name
from biscuit.common.icons import Icons
from biscuit.common.ui import Scrollbar

from ..panelview import PanelView


class Logs(PanelView):
    """The Logs view.

    The Logs view displays the logs of the application.
    - Show info, warning, error, trace logs.
    - Clear all logs.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.__actions__ = ((Icons.CLEAR_ALL, self.clear_all),)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.queue = queue.Queue()

        self.text = tk.Text(
            self,
            relief=tk.FLAT,
            padx=10,
            pady=10,
            font=("Consolas", 11),
            **self.base.theme.editors.text,
        )
        self.text.grid(row=0, column=0, sticky=tk.NSEW)

        self.scrollbar = Scrollbar(self, style="EditorScrollbar")
        self.scrollbar.grid(sticky=tk.NSEW, row=0, column=1)

        self.text.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.text.yview)

        fontbold = ("Consolas", 11, "bold")

        self.text.tag_config("time", foreground=self.base.theme.views.panel.logs.time)
        self.text.tag_config(
            "caller", foreground=self.base.theme.views.panel.logs.caller
        )

        self.text.tag_config("info", foreground=self.base.theme.views.panel.logs.info)
        self.text.tag_config(
            "warning", foreground=self.base.theme.views.panel.logs.warning
        )
        self.text.tag_config(
            "error", foreground=self.base.theme.views.panel.logs.error, font=fontbold
        )
        self.text.tag_config("trace", foreground=self.base.theme.border)

        self.gui_refresh_loop()

    def gui_refresh_loop(self) -> None:
        if not self.queue.empty():
            self.write(*self.queue.get())

        self.after(10, self.gui_refresh_loop)

    def write(self, *args) -> None:
        self.text.config(state=tk.NORMAL)
        for i in args:
            if isinstance(i, tuple):
                self.text.insert(tk.END, i[0], i[1])
            else:
                self.text.insert(tk.END, i)
        self.text.config(state=tk.DISABLED)
        self.text.see(tk.END)

    def newline(self) -> None:
        self.queue.put(("\n",))

    def log(self, type: str, caller: str, text: str) -> None:
        self.queue.put(
            (
                "[",
                (datetime.now().strftime("%H:%M:%S:%f"), "time"),
                "]",
                type,
                "[",
                (caller, "caller"),
                f"]: {text}",
            )
        )
        self.newline()

    def info(self, text: str) -> None:
        """info level log"""

        self._std_log(" [info] ", "info", text)

    def warning(self, text: str) -> None:
        """warning level log"""

        self._std_log(" [warning] ", "warning", text)

    def error(self, text: str) -> None:
        """error level log"""

        self._std_log(" [error] ", "error", text)

    def trace(self, text: str) -> None:
        """trace level log"""

        # traces are fully greyed out
        self.queue.put(
            (
                (
                    f"[{datetime.now().strftime('%H:%M:%S:%f')}] [trace] [{caller_class_name(skip=3)}]: {text}",
                    "trace",
                ),
                "\n",
            )
        )

    def _std_log(self, kindtext: str, kind: str, text: str) -> None:
        self.log((kindtext, kind), caller_class_name(skip=3), text)

    def rawlog(self, text: str, kind: int):
        match kind:
            case 1:
                self.error(text)
            case 2:
                self.warning(text)
            case 3:
                self.info(text)
            case _:
                self.trace(text)

    def clear_all(self, *_) -> None:
        self.text.config(state=tk.NORMAL)
        self.text.delete(1.0, tk.END)
        self.text.config(state=tk.DISABLED)



from dataclasses import dataclass
from typing import List


@dataclass
class Change:
    start: List[int]
    old_end: List[int]
    new_end: List[int]
    old_text: str
    new_text: str

    def update(self, start, old_end, new_end, old_text, new_text):
        self.start = start
        self.old_end = old_end
        self.new_end = new_end
        self.old_text = old_text
        self.new_text = new_text


# @dataclass
# class Changes:
#     change_list: List[Change]



from __future__ import annotations

import typing

from .menu import Menu

if typing.TYPE_CHECKING:
    from biscuit import App


class TextEditorContextMenu(Menu):
    def __init__(self, master: App, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        commands = self.base.commands

        self.e = None

        self.add_command("Cut", commands.cut)
        self.add_command("Copy", commands.copy)
        self.add_command("Paste", commands.paste)

        self.add_separator()

        self.add_command("Rename Symbol", commands.rename_symbol)
        self.add_command("Toggle Comment", commands.toggle_comment)
        # self.add_command("Format Document", commands.format_document)
        # self.add_command("Format Selection", commands.format_selection)

        self.add_separator()

        self.add_command("Go to Definition", commands.go_to_symbol_definition)
        self.add_command("Go to References", commands.find_symbol_references)
        # self.add_command("Go to Declaration", commands.go_to_declaration)
        # self.add_command("Go to Type Definition", commands.go_to_type_definition)
        self.add_command("Go to Line/Column...", commands.goto_line_column)
        self.add_command("Go to Symbol in Editor", commands.show_symbol_palette)

        menu = self.add_menu("Selection").menu
        menu.add_command("Select All", commands.select_all)
        menu.add_command("Select Line", commands.select_line)
        menu.add_command("Delete Line", commands.delete_line)
        menu.add_separator()
        menu.add_command("Copy Line Up", commands.copy_line_up)
        menu.add_command("Copy Line Down", commands.copy_line_down)
        menu.add_command("Move Line Up", commands.move_line_up)
        menu.add_command("Move Line Down", commands.move_line_down)
        menu.add_command("Duplicate Selection", commands.duplicate_selection)

        self.add_separator()

        self.add_command("Command Palette...", commands.show_command_palette)

    def get_coords(self, e) -> list:
        return e.x_root, e.y_root

    def show(self, *e) -> None:
        self.e = e
        super().show(*e)



from __future__ import annotations

import tkinter as tk
import typing
from tkinter import ttk
from tkinter.font import Font

if typing.TYPE_CHECKING:
    from .settings import Settings


class Style(ttk.Style):
    """Handles the styling of the app"""

    def __init__(self, settings: Settings, *args, **kwargs) -> None:
        super().__init__(settings.base, *args, **kwargs)
        self.settings = settings
        self.base = settings.base
        self.theme = settings.config.theme

        self.configure(
            "TCheckbutton",
            background=self.theme.editors.background,
            font=self.settings.uifont,
        )
        self.configure("TFrame", background=self.theme.editors.background)
        self.gen_fileicons()

        self.config_treeview()
        self.config_tree_scrollbar()

    def config_treeview(self) -> None:

        self.configure(
            "Treeview.treearea",
            font=self.settings.uifont,
            relief="flat",
            rowheight=25,
            highlightthickness=0,
            bd=0,
            **self.theme.utils.tree.item,
            padding=0,
        )
        self.configure(
            "Treeview",
            font=self.settings.uifont,
            relief="flat",
            rowheight=25,
            highlightthickness=0,
            bd=0,
            **self.theme.utils.tree.item,
            padding=0,
        )

        self.layout("Treeview", [("Treeview.treearea", {"sticky": "nswe"})])

        self.monofont = Font(family=self.settings.config.font[0], size=10)
        self.layout("mono.Treeview", [("Treeview.treearea", {"sticky": "nswe"})])
        self.configure(
            "mono.Treeview.treearea",
            font=self.monofont,
            relief="flat",
            rowheight=25,
            highlightthickness=0,
            bd=0,
            **self.theme.utils.tree.item,
            padding=0,
        )
        self.configure(
            "mono.Treeview",
            font=self.monofont,
            relief="flat",
            rowheight=25,
            highlightthickness=0,
            bd=0,
            **self.theme.utils.tree.item,
            padding=0,
        )

        self.layout("secondary.Treeview", [("Treeview.treearea", {"sticky": "nswe"})])
        self.configure(
            "secondary.Treeview.treearea",
            font=self.monofont,
            relief="flat",
            rowheight=25,
            highlightthickness=0,
            bd=0,
            **self.theme.utils.secondary_tree.item,
            padding=0,
        )
        self.configure(
            "secondary.Treeview",
            font=self.monofont,
            relief="flat",
            rowheight=25,
            highlightthickness=0,
            bd=0,
            **self.theme.utils.secondary_tree.item,
            padding=0,
        )

    def config_tree_scrollbar(self) -> None:
        self.element_create("TreeScrollbar.trough", "from", "clam")
        self.element_create("TreeScrollbar.thumb", "from", "clam")

        self.layout(
            "TreeScrollbar",
            [
                (
                    "TreeScrollbar.trough",
                    {
                        "sticky": "ns",
                        "children": [
                            ("TreeScrollbar.thumb", {"unit": "1", "sticky": "nsew"})
                        ],
                    },
                )
            ],
        )

        bg, _, highlight, _ = self.theme.utils.scrollbar.values()
        self.configure(
            "TreeScrollbar",
            gripcount=0,
            background=bg,
            troughcolor=bg,
            bordercolor=bg,
            lightcolor=bg,
            darkcolor=bg,
            arrowsize=14,
        )
        self.map(
            "TreeScrollbar",
            background=[("pressed", highlight), ("!disabled", self.theme.border)],
        )

        self.element_create("EditorScrollbar.trough", "from", "clam")
        self.element_create("EditorScrollbar.thumb", "from", "clam")

        self.layout(
            "EditorScrollbar",
            [
                (
                    "EditorScrollbar.trough",
                    {
                        "sticky": "nsew",
                        "children": [("EditorScrollbar.thumb", {"sticky": "nsew"})],
                    },
                )
            ],
        )
        self.configure(
            "EditorScrollbar",
            gripcount=0,
            background=bg,
            troughcolor=bg,
            bordercolor=bg,
            lightcolor=bg,
            darkcolor=bg,
        )
        self.map(
            "EditorScrollbar",
            background=[("pressed", highlight), ("!disabled", self.theme.border)],
        )

    def gen_fileicons(self) -> None:
        self.document_icn = tk.PhotoImage(
            "document",
            data="""
        iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAAJ2AAACdgBx6C5rQA
        AABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoAAADlSURBVDiNpZGxTgJBFEXPW9aGCRTYWht+wyVEEuq1md
        7Exm/A2NjxFUvBD1CQ7JZWlOhXQCNsoYnPajZk3Zks4VaTmfvOO8nAhRF3SJfa2X2XLwi39ZIi74XtzoOA0eIwUZVVQ+cDu
        AHmuTWz+mNUbdGo57HcKiTAc5KVb15AKIU1G4Ux6GMd0grgICIyBX26yw737j5uMZsm2VEBVBUAIeqfbeDLP4PcGmkqujgb
        LyDJjsuLDAJJWwFyax6ainV1L8BX9KX6BZHfr7ZDp93KYBCb9f6nfFUYhoZV+by+MutzLIP5A16TRi/mS3m5AAAAAElFTkS
        uQmCC
        """,
        )

        self.folder_icn = tk.PhotoImage(
            "foldericon",
            data="""
        iVBORw0KGgoAAAANSUhEUgAAABAAAAAMCAYAAABr5z2BAAAACXBIWXMAAA7DAAAOwwHHb6hkAAAAGXRFWHRTb2Z0d2FyZQB
        3d3cuaW5rc2NhcGUub3Jnm+48GgAAAJBJREFUKJHdzTEKwkAUhOF/loCFRbAVr+IhLAWLCPaW3sFGPIOm1Bt4hxSSEwRs7Z
        UdayErmnROO++bp93htJK0BUa8pxEq1ovZhQ/R/ni+G/LWEjW2y4Stx4NnmUU7l9R6YTxBbFLfb49sGlL4m9ieh84aAA17D
        sCfDLiHdwDqrlpwDTHGAqiA+IONQIW0fAFkySdEGFdeCgAAAABJRU5ErkJggg==""",
        )



import tkinter as tk

from biscuit import __version__
from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconButton, IconLabelButton


class SearchBar(Frame):
    """Search bar for the menu bar.

    - When no active directory, opens the command palette, shows the biscuit version
    - When active directory, opens the file search, shows the active directory name"""

    def __init__(self, master, *args, **kwargs):
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)

        self.label = IconLabelButton(
            self,
            text=f"Biscuit {__version__}",
            icon=Icons.SEARCH,
            padx=150,
            callback=self.onclick,
        )
        font = self.base.settings.uifont.copy()
        font.config(size=8)
        self.label.text_label.config(font=font)
        self.label.icon_label.config(font=("codicon", 10))
        self.label.pack(side=tk.LEFT, fill=tk.BOTH, expand=True, pady=1, padx=1)

        self.palette_button = IconButton(
            self,
            icon=Icons.TERMINAL,
            event=self.base.commands.show_command_palette,
            iconsize=10,
        )
        self.palette_button.config(**self.base.theme.utils.iconlabelbutton)
        self.palette_button.pack(
            side=tk.RIGHT, fill=tk.BOTH, expand=True, pady=1, padx=(0, 1)
        )

    def onclick(self, *_):
        if self.base.active_directory:
            self.base.commands.search_files()
        else:
            self.base.commands.show_command_palette()



from __future__ import annotations

import tkinter as tk
import typing

from ..actionset import ActionSet
from ..ui import Frame, Toplevel
from .item import PaletteItem
from .searchbar import SearchBar

if typing.TYPE_CHECKING:
    from biscuit import App


# TODO: enlarge current item, add shortcuts, secondary text
class Palette(Toplevel):
    """Palette

    Palette is an action menu centered horizontally and aligned to top of root.
    It is used to display a list of actions that can be performed by the user.
    The user can search for actions and select them using the keyboard or mouse.

    Actions are registered as ActionSets, which are accessed with prefixes from search bar.
    Palette supports setting pinned actions that are displayed at the top of the palette.

    When no prefix is detected, palette turns file search mode on.
    Help is displayed when the user types '?' in the search bar.

    Palette can also be used to take input from the user, e.g: GitHub clone URL, go-to line number.
    """

    def __init__(self, master: App, width=80, *args, **kwargs) -> None:
        """Initializes the Palette

        Args:
            master (App): The main application instance
            width (int, optional): The width of the palette. Defaults to 80."""

        super().__init__(master, *args, **kwargs)
        self.config(pady=2, padx=2, bg=self.base.theme.border)

        self.container = Frame(self, **self.base.theme.palette, padx=2, pady=2)
        self.container.pack(fill=tk.BOTH, expand=True)

        self.width = round(width * self.base.scale)
        self.active = False

        self.withdraw()
        self.overrideredirect(True)

        self.container.grid_columnconfigure(0, weight=1)
        self.container.grid_rowconfigure(0, weight=1)

        self.row = 1
        self.selected = 0
        self.start_index = 0

        self.shown_items = []

        self.actionsets = []
        self.active_set = None
        self.active_items = None

        self.searchbar = SearchBar(self)
        self.searchbar.grid(row=0, sticky=tk.EW, in_=self.container, pady=(0, 2))

        self.configure_bindings()

    def register_actionset(self, actionset_ref: ActionSet) -> None:
        """Registers an actionset to the palette
        NOTE: actionset_ref is a lambda returning the actionset instead of the actionset itself

        Args:
            actionset_ref (ActionSet): lambda returning the actionset instead of the actionset itself
        """
        self.actionsets.append(actionset_ref)

    def generate_help_actionset(self) -> None:
        """Generates the help actionset

        The help actionset is generated from the registered actionsets.
        It is displayed when the user types '?' in the search bar.
        Make sure this is called after all actionsets have been registered."""

        self.help_actionset = ActionSet("Help", "?")
        for i in self.actionsets:
            i = i()  # get the actionset
            if i.prefix:
                self.help_actionset.append(
                    (
                        i.prefix,
                        lambda _, i=i: self.after(50, self.show, i.prefix),
                        i.description,
                    )
                )

        # print([i() for i in self.actionsets])
        self.register_actionset(lambda: self.help_actionset)

    def add_item(self, text: str, command, *args, **kwargs) -> PaletteItem:
        """Adds an item to the palette

        Args:
            text (str): The text to display on the item
            command (str): The command to execute when the item is selected

        Returns:
            PaletteItem: The item that was added to the palette"""

        new_item = PaletteItem(self, text, command, *args, **kwargs)
        new_item.grid(row=self.row, sticky=tk.EW, in_=self.container)

        self.shown_items.append(new_item)

        self.row += 1
        self.refresh_selected()
        return new_item

    def configure_bindings(self) -> None:
        self.bind("<FocusOut>", self.hide)
        self.bind("<Escape>", self.hide)

        self.row += 1
        self.refresh_selected()

    def reset_start_index(self) -> None:
        self.start_index = 0

    def on_mousewheel(self, event) -> str:
        if not self.active_items:
            return "break"

        # start_index must be between 0 and len(active_items) - 10
        self.start_index = min(
            max(0, self.start_index - event.delta // 120), len(self.active_items) - 10
        )
        self.show_items(self.active_items)

    def pick_actionset(self, actionset: ActionSet) -> None:
        """Picks an actionset to display in the palette

        Args:
            actionset (ActionSet): The actionset to display in the palette"""

        self.active_set = actionset

    def pick_file_search(self, term: str) -> None:
        """Picks the file search actionset

        Args:
            term (str): The search term"""

        self.active_set = self.base.explorer.get_actionset(term)

    def choose(self, *_) -> None:
        """Executes the selected item's command

        If an item is selected, the command is executed with the search term as an argument.
        """

        if item := self.shown_items[self.selected]:
            picked_command = item.command
            term = self.searchbar.term

            self.hide()
            picked_command(term)

    def get_items(self) -> ActionSet:
        return self.active_set

    def hide(self, *args) -> None:
        """Hides the palette"""

        self.withdraw()
        self.reset()

        self.container.unbind_all("<MouseWheel>")

    def hide_all_items(self) -> None:
        """Hides all items in the palette"""

        for i in self.shown_items:
            i.destroy()

        self.shown_items = []
        self.row = 1

    def reset_selection(self) -> None:
        """Resets the selected item to the first item in the palette"""

        self.selected = 0
        self.refresh_selected()

    def refresh_selected(self) -> None:
        """Refreshes the selected item"""

        if not self.shown_items:
            return

        for i in self.shown_items:
            i.deselect()

        try:
            self.shown_items[self.selected].select()
        except IndexError as e:
            self.base.logger.error(f"Command '{self.selected}' doesnt exist: {e}")

    def reset(self) -> None:
        """Resets the palette

        Clears the search bar and resets the selection."""

        self.searchbar.clear()
        self.reset_selection()

    def search_bar_enter(self, *_) -> str:
        """Handles the enter key press in the search bar"""

        self.choose()
        return "break"

    def show_no_results(self) -> None:
        """Shows a 'No results found' message in the palette"""

        self.hide_all_items()
        self.reset_selection()
        self.add_item("No results found", lambda _: ...)

    def select(self, delta: int) -> None:
        """Selects an item in the palette

        Args:
            delta (int): The change in selection"""

        self.selected += delta
        self.selected = min(max(0, self.selected), len(self.shown_items) - 1)
        self.refresh_selected()

    def show_items(self, items: list[PaletteItem]) -> None:
        """Shows a list of items in the palette

        Args:
            items (list[PaletteItem]): The items to display in the palette"""

        self.hide_all_items()
        self.active_items = items

        for i in self.active_items[self.start_index : self.start_index + 10]:
            item = self.add_item(*i)
            item.mark_term(self.searchbar.term)

        self.reset_selection()

    def show(self, prefix: str = None, default: str = None) -> None:
        """Shows the palette

        Args:
            prefix (str, optional): The prefix to search for. Defaults to None.
            default (str, optional): The default search term. Defaults to None."""

        self.update_idletasks()
        self.update()

        x = self.master.winfo_rootx() + int(
            (self.master.winfo_width() - self.winfo_width()) / 2
        )
        y = self.master.winfo_rooty() + int((self.master.winfo_height() / 2) - 200)
        self.geometry(f"+{x}+{y}")
        self.deiconify()

        self.focus_set()
        self.searchbar.focus()
        self.searchbar.add_prefix(prefix)

        if default:
            self.searchbar.set_search_term(default)

        self.container.bind_all("<MouseWheel>", self.on_mousewheel)



"""ANSI escape sequence utilities."""

import re

SEQ = re.compile(r"\x1b(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
NEWLINE = re.compile(r"\x1b\[\d+\;1H")


def strip_ansi_escape_sequences(string):
    return SEQ.sub("", string)


def replace_newline(string):
    return NEWLINE.sub("\n", string)



from .endpoint import Endpoint


class Commands(Endpoint):
    """Commands endpoint

    This class is used to interact with the commands API.
    Commands are actions that can be executed by the user.
    Command Palette is the main way to interact with commands.
    """

    def __init__(self, *a) -> None:
        super().__init__(*a)
        self.settings = self.base.settings

        self.register_command = self.settings.register_command

    @property
    def commands(self) -> None:
        """Return all registered commands"""

        return self.settings.commands.copy()



import tkinter as tk

import git


def custom_get(start, end):
    content = text.get(start, end)
    tag_ranges = text.tag_ranges("codelens")

    for tag_start, tag_end in zip(tag_ranges[0::2], tag_ranges[1::2]):
        content = content.replace(text.get(tag_start, tag_end), '')

    return content

def get_commit_data(file_path):
    repo = git.Repo(search_parent_directories=True)
    commits = repo.iter_commits(paths=file_path)
    commit_data = [commit.message.strip() for commit in commits]
    return commit_data

def handle_codelens_click(commit):
    # print(f"Commit data: {commit}")
    print(custom_get('1.0', tk.END))

root = tk.Tk()

text = tk.Text(root)
text.pack()

with open("SECURITY.md") as f:
    content = f.read()
    text.insert(tk.END, content)

# Create CodeLens annotations with commit data
commit_data = get_commit_data("SECURITY.md")
for line_number, commit in enumerate(commit_data, start=1):
    text.tag_configure("codelens", underline=True, foreground="blue")
    text.tag_bind("codelens", "<Button-1>", lambda _, commit=commit: handle_codelens_click(commit))
    text.insert(f"{line_number}.0", f"[CodeLens] {commit}\n", "codelens")


root.mainloop()



from __future__ import annotations

import os
import tkinter as tk
import typing

from biscuit.common import Icons
from biscuit.common.ui import Frame, Icon, IconButton

if typing.TYPE_CHECKING:
    from biscuit.editor import Editor

    from .editorsbar import EditorsBar

# TODO: show modified, saved state in the tab


class Tab(Frame):
    """Editor Tab

    An editor instance is attached to each tab.
    Shows the filename, icon and close button.
    """

    def __init__(self, master: EditorsBar, editor: Editor, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master: EditorsBar = master
        self.editor = editor
        self.selected = False

        self.bg, self.fg, self.hbg, self.hfg = (
            self.base.theme.layout.content.editors.bar.tab.values()
        )
        self.config(bg=self.bg)

        self.icon = Icon(
            self,
            Icons.FILE,
            iconsize=12,
            **self.base.theme.layout.content.editors.bar.tab.icon,
        )
        self.icon.pack(side=tk.LEFT, padx=5, fill=tk.Y)

        self.name = tk.Label(
            self,
            text=(
                f"{editor.filename} (working tree)" if editor.diff else editor.filename
            ),
            padx=5,
            font=self.base.settings.uifont,
            bg=self.bg,
            fg=self.fg,
        )
        self.name.pack(side=tk.LEFT)

        self.closebtn = IconButton(
            self,
            Icons.CLOSE,
            iconsize=12,
            event=self.close,
        )
        self.closebtn.config(**self.base.theme.layout.content.editors.bar.tab.close)
        self.closebtn.pack(side=tk.RIGHT, fill=tk.Y)

        self.bind("<Button-1>", self.select)
        self.name.bind("<Button-1>", self.select)

        self.bind("<Enter>", self.on_hover)
        self.bind("<Leave>", self.off_hover)

    def close(self, *_) -> None:
        self.master.close_tab(self)

    def apply_color(self, color: str) -> None:
        self.icon.config(bg=color)
        self.name.config(bg=color)
        self.config(bg=color)
        self.closebtn.config(bg=color)

    def on_hover(self, *_) -> None:
        if not self.selected:
            self.apply_color(self.hbg)
            self.closebtn.config(activeforeground=self.hfg, fg=self.fg)
            self.hovered = True

    def off_hover(self, *_) -> None:
        if not self.selected:
            self.apply_color(self.bg)
            self.closebtn.config(activeforeground=self.fg, fg=self.bg)
            self.hovered = False

    def deselect(self, *_) -> None:
        if self.selected:
            self.editor.grid_remove()
            self.apply_color(self.bg)
            self.closebtn.config(activeforeground=self.fg, fg=self.bg)
            self.selected = False

    def select(self, *_) -> None:
        if not self.selected:
            self.master.set_active_tab(self)
            if self.base.active_directory and self.editor.filename:
                self.base.set_title(
                    f"{self.editor.filename} - {os.path.basename(self.base.active_directory)}"
                )
            elif self.editor.filename:
                self.base.set_title(self.editor.filename)
            self.editor.grid(column=0, row=1, sticky=tk.NSEW, in_=self.master.master)

            self.apply_color(self.hbg)
            self.closebtn.config(activeforeground=self.hfg, fg=self.fg)
            self.selected = True

        if (
            self.editor.path
            and self.editor.exists
            and self.editor.showpath
            and not self.editor.diff
        ):
            self.master.show_breadcrumbs()
            self.base.breadcrumbs.set_path(self.editor.path)
        else:
            self.master.hide_breadcrumbs()



"""
Windows specific DPI based scaling for tkinter.
NOTE: Currently not in use.
"""

import re


def Get_HWND_DPI(window_handle):
    # To detect high DPI displays and avoid need to set Windows compatibility flags
    from ctypes import windll, pointer, wintypes

    try:
        windll.shcore.SetProcessDpiAwareness(1)
    except Exception:
        pass
    DPI100pc = 96  # DPI 96 is 100% scaling
    DPI_type = 0   # MDT_EFFECTIVE_DPI = 0, MDT_ANGULAR_DPI = 1, MDT_RAW_DPI = 2
    winH = wintypes.HWND(window_handle)
    monitorhandle = windll.user32.MonitorFromWindow(
        winH, wintypes.DWORD(2)
    )  # MONITOR_DEFAULTTONEAREST = 2
    X = wintypes.UINT()
    Y = wintypes.UINT()
    try:
        windll.shcore.GetDpiForMonitor(
            monitorhandle, DPI_type, pointer(X), pointer(Y)
        )
        return X.value, Y.value, (X.value + Y.value) / (2 * DPI100pc)
    except Exception:
        return 96, 96, 1  # Assume standard Windows DPI & scaling


def TkGeometryScale(s, cvtfunc):
    patt = r"(?P<W>\d+)x(?P<H>\d+)\+(?P<X>\d+)\+(?P<Y>\d+)"  # format "WxH+X+Y" for tk
    R = re.compile(patt).search(s)
    G = str(cvtfunc(R.group("W"))) + "x"
    G += str(cvtfunc(R.group("H"))) + "+"
    G += str(cvtfunc(R.group("X"))) + "+"
    G += str(cvtfunc(R.group("Y")))
    return G


def MakeTkDPIAware(TKGUI):
    TKGUI.DPI_X, TKGUI.DPI_Y, TKGUI.DPI_scaling = Get_HWND_DPI(TKGUI.winfo_id())
    TKGUI.TkScale = lambda v: int(float(v) * TKGUI.DPI_scaling)
    TKGUI.TkGeometryScale = lambda s: TkGeometryScale(s, TKGUI.TkScale)




from .endpoint import Endpoint


class Editors(Endpoint):
    def __init__(self, *a) -> None:
        super().__init__(*a)
        self.editors = self.base.editorsmanager
        self.theme = self.base.theme

    def add_editor(self, editor) -> None:
        self.editors.add_editor(editor)

    def open_editor(self, path=None, exists=True) -> None:
        self.editors.open_editor(path, exists)

    def open_diff_editor(self, path, exists) -> None:
        self.editors.open_diff_editor(path, exists)

    def refresh(self) -> None:
        self.editors.refresh()

    def active_editor(self):
        return self.editors.active_editor

    def set_active_editor(self, editor) -> None:
        self.editors.set_active_editor(editor)

    def close_editor(self, editor) -> None:
        self.editors.close_editor(editor)

    def close_active_editor(self) -> None:
        self.editors.close_active_editor()



import tkinter as tk
import tkinter.ttk as ttk
import typing

from .native import Frame
from .scrollbar import Scrollbar


class Tree(Frame):
    """Treeview with an auto hiding scrollbar

    Args:
        master (tk.Widget): Parent widget
        startpath (str, optional): Starting path for the tree.
        doubleclick (function, optional): Function to call on double click. Defaults to lambda _: None.
        singleclick (function, optional): Function to call on single click. Defaults to lambda _: None.
        columns (tuple, optional): Columns to display. Defaults to ("fullpath", "type").
    """

    def __init__(
        self,
        master,
        startpath="",
        doubleclick=lambda _: None,
        singleclick=lambda _: None,
        columns=("fullpath", "type"),
        style="",
        *args,
        **kwargs,
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.utils.tree)

        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

        self.path = startpath
        self.doubleclick = doubleclick
        self.singleclick = singleclick

        self.tree = ttk.Treeview(
            self,
            show="tree",
            columns=columns,
            displaycolumns="",
            style=style,
            selectmode=tk.BROWSE,
        )
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)

        self.scrollbar = Scrollbar(
            self, orient=tk.VERTICAL, command=self.tree.yview, style="TreeScrollbar"
        )
        self.scrollbar.grid(row=0, column=1, sticky=tk.NS)

        self.tree.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.tree.yview)

        self.bind("<Double-Button-1>", self.doubleclick)
        self.bind("<<TreeviewSelect>>", self.check_singleclick)

        # self.tree.bind("<Motion>", self.on_motion)
        # self.tree.bind("<Leave>", self.on_leave)
        # self.tree.tag_configure("hover", background=self.base.theme.border)
        # self.hovered = None

    def on_motion(self, event):
        try:
            self.item(self.hovered, tags=())
        except tk.TclError:
            pass

        item = self.identify_row(event.y)
        try:
            self.item(item, tags=("hover",))
            self.hovered = item
        except tk.TclError:
            pass

    def on_leave(self, _):
        if self.hovered:
            try:
                self.item(self.hovered, tags=())
            except tk.TclError:
                pass
            self.hovered = None

    def bind(self, *args, **kwargs) -> None:
        self.tree.bind(*args, **kwargs)

    def check_singleclick(self, _) -> None:
        try:
            if self.item_type(self.focus()) == "file":
                if self.singleclick:
                    self.singleclick(self.item_fullpath(self.focus()))
            else:
                self.toggle_node(self.focus())
        except:
            self.toggle_node(self.focus())

    def clear_node(self, node) -> None:
        self.tree.delete(*self.tree.get_children(node))

    def clear_tree(self) -> None:
        self.clear_node("")

    def collapse_all(self) -> None:
        for node in self.tree.get_children():
            self.tree.item(node, open=False)

    def delete(self, *a, **kw) -> None:
        self.tree.delete(*a, *kw)

    def focus(self, *args) -> str:
        return self.tree.focus(*args) or ""

    def get_children(self, *a, **kw) -> str:
        return self.tree.get_children(*a, **kw)

    def identify_row(self, y) -> str:
        return self.tree.identify_row(y)

    def insert(self, *args, **kwargs):
        try:
            return self.tree.insert(*args, **kwargs)
        except tk.TclError:
            return None

    def add(self, *a, **kw):
        return self.tree.insert("", "end", *a, **kw)

    def is_open(self, node):
        return self.tree.item(node, "open")

    def item(self, *a, **kw):
        return self.tree.item(*a, **kw)

    def item_type(self, item):
        return self.set(item, "type")

    def item_fullpath(self, item):
        return self.set(item, "fullpath")

    def parent(self, *args, **kwargs):
        return self.tree.parent(*args, **kwargs)

    def parent_selected(self):
        return self.parent(self.focus())

    def selected_parent_path(self):
        return self.item_fullpath(self.parent_selected())

    def selection(self, *args, **kwargs):
        return self.tree.selection(*args, **kwargs)

    def selection_set(self, *args, **kwargs):
        return self.tree.selection_set(*args, **kwargs)

    def selected_path(self):
        return self.item_fullpath(self.focus())

    def selected_type(self):
        return self.item_type(self.focus())

    def is_file_selected(self):
        return self.selected_type() == "file"

    def is_directory_selected(self):
        return self.selected_type() == "directory"

    def set(self, *args, **kwargs):
        return self.tree.set(*args, **kwargs)

    def __getattr__(self, name) -> typing.Any:
        """For methods that are not in this class, assume it is present in tree widget
        and delegate the call to the `tree` widget."""

        text_methods = set(dir(self.tree))
        if name in text_methods:
            attr = getattr(self.tree, name)
            return attr
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{name}'"
        )

    def toggle_node(self, node) -> None:
        try:
            if self.item_type(node) == "directory":
                if self.is_open(node):
                    self.tree.item(node, open=False)
                else:
                    self.tree.item(node, open=True)
                    self.tree.event_generate("<<Open>>")
        except:
            if self.is_open(node):
                self.tree.item(node, open=False)
            else:
                self.tree.item(node, open=True)



from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconButton


class BaseGame(Frame):
    """Base class for games.

    Games may inherit from this class and implement the following methods:
    - `__init__`
    - `reload`

    Otherwise make sure to set the following attributes:
    - `path` (str): path to the game file
    - `filename` (str): name of the game file
    - `exists` (bool): whether the game file exists
    - `diff` (bool): whether the game file has been modified
    - `showpath` (bool): whether to display the path of the game file
    - `editable` (bool): whether the game file is editable
    - `__buttons__` (list): list of buttons to be added to the editor bar

    and implement the following methods:
    - `reload`: reload the game file
    - `save`: save the game file
    """

    def __init__(self, master, path=None, *args, **kwargs) -> None:
        """Initializes the game frame

        Args:
            master (tk.Tk): master window
            path (str): path to the game file
        """

        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.editors)
        self.path = path
        self.filename = None

        self.exists = False
        self.showpath = False
        self.diff = False
        self.editable = False

        self.__buttons__ = ((Icons.REFRESH, self.reload),)

    def add_buttons(self, icon, event) -> None:
        """Adds a button to the editor bar

        Args:
            icon (str): name of the icon
            event (function): function to be called on button press
        """

        self.__buttons__.append((icon, event))

    def create_buttons(self, editorbar) -> None:
        """Not to be called directly. Creates buttons for the editor bar"""

        self.__buttons__ = [
            IconButton(editorbar, *button) for button in self.__buttons__
        ]

    def reload(self, *_) -> None:
        """Reloads the game
        This method should be implemented by the subclass"""

        ...

    def save(self, *_) -> None:
        """Saves the game
        This method should be implemented by the subclass"""

        ...



import tkinter as tk

from biscuit.common.ui import Frame

from .sidebar_item_toolbar import SideBarItemToolBar


class SideBarViewItem(Frame):
    """Container for the sidebar view items.

    - Contains the itembar and the content of the item.
    """

    __actions__ = []
    title = "Item"

    def __init__(
        self, master, title: str = None, itembar=True, *args, **kwargs
    ) -> None:
        """Initialize the sidebar view item

        Args:
            master (tk.Tk): root window
            title (str, optional): title of the item. Defaults to None.
            itembar (bool, optional): whether to show the itembar. Defaults to True."""

        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.views.sidebar.item)

        if title:
            self.title = title

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(1, weight=1)

        self.enabled = True
        self.itembar_enabled = itembar

        if itembar:
            self.itembar = SideBarItemToolBar(self, self.title, self.__actions__)
            self.itembar.grid(row=0, column=0, sticky=tk.NSEW)

        self.content = Frame(self, **self.base.theme.views.sidebar.item)
        self.content.master = self.master
        self.content.grid_rowconfigure(0, weight=1)
        self.content.grid_columnconfigure(0, weight=1)
        self.content.grid(row=1 if itembar else 0, column=0, sticky=tk.NSEW)

    def add_action(self, icon: str, event) -> None:
        if self.itembar_enabled:
            self.itembar.add_action(icon, event)

    def set_title(self, title: str) -> None:
        if self.itembar_enabled:
            self.itembar.set_title(title)

    def toggle(self, *_) -> None:
        if not self.enabled:
            self.enable()
        else:
            self.disable()

    def enable(self) -> None:
        if not self.enabled:
            self.content.grid()
            self.enabled = True

    def disable(self) -> None:
        if self.enabled:
            self.content.grid_remove()
            self.enabled = False



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconButton


class SideBarItemToolBar(Frame):
    def __init__(self, master, title: str, buttons=(), *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.views.sidebar.itembar)

        self.grid_columnconfigure(1, weight=1)

        self.title = tk.StringVar(self)
        if title:
            self.set_title(title)

        self.toggle = IconButton(
            self, icon=Icons.CHEVRON_DOWN, event=self.toggle_content, width=1
        )
        self.toggle.grid(row=0, column=0)

        self.label_title = tk.Label(
            self,
            anchor=tk.W,
            textvariable=self.title,
            font=self.base.settings.uifont,
            **self.base.theme.views.sidebar.itembar.title
        )
        self.label_title.grid(row=0, column=1, sticky=tk.EW)

        self.buttoncolumn = 0
        self.actions = Frame(self, **self.base.theme.views.sidebar.itembar)
        self.actions.grid(row=0, column=2, sticky=tk.E)

        self.add_actions(buttons)

    def add_action(self, *args) -> None:
        IconButton(self.actions, *args).grid(row=0, column=self.buttoncolumn)
        self.buttoncolumn += 1

    def add_actions(self, buttons) -> None:
        for btn in buttons:
            self.add_action(*btn)

    def set_title(self, title: str) -> None:
        self.title.set(title.upper())

    def toggle_content(self, *_) -> None:
        if not self.master.enabled:
            self.toggle.set_icon(Icons.CHEVRON_DOWN)
        else:
            self.toggle.set_icon(Icons.CHEVRON_RIGHT)
        self.master.toggle()

    def hide_content(self, *_) -> None:
        self.toggle.set_icon(Icons.CHEVRON_RIGHT)
        self.master.disable()

    def show_content(self, *_) -> None:
        self.toggle.set_icon(Icons.CHEVRON_DOWN)
        self.master.enable()



import tkinter as tk

from biscuit.common.icons import Icons

from ..sidebar_item import SideBarViewItem
from .stageditem import StagedChangeItem


class StagedChanges(SideBarViewItem):
    """The Staged Changes view.

    The Staged Changes view allows the user to view the staged changes made to the active document.
    - Remove staged changes.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        self.__actions__ = ((Icons.REMOVE, self.git_remove_all),)
        self.title = "Staged Changes"
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.views.sidebar.item)
        self.items = {}

    def refresh(self) -> None:
        if not self.items:
            self.itembar.hide_content()
        else:
            self.itembar.show_content()

    def clear_tree(self, *_) -> None:
        for item in self.items.values():
            try:
                item.destroy()
            except:
                pass

        self.items.clear()

    def clear(self, otherthan: list = []) -> None:
        if not otherthan:
            return self.clear_tree()

        for path in list(self.items.keys()):
            if path not in otherthan:
                self.items[path].destroy()
                del self.items[path]

    def add_item(self, path, kind) -> None:
        if path in self.items:
            return

        new_item = StagedChangeItem(self.content, path, kind)
        new_item.master = self
        new_item.pack(fill=tk.X)
        self.items[path] = new_item

    def git_remove_all(self, *_) -> None:
        if staged := list(self.items.keys()):
            self.base.git.repo.unstage_files(*staged)
            self.master.master.open_repo()



class TextIndex:
    """A class that represents tkinter text widget indexes."""

    def __init__(self, index):
        self.line, self.column = map(int, index.split("."))

    def __lt__(self, other):
        if self.line == other.line:
            return self.column < other.column
        return self.line < other.line

    def __le__(self, other):
        return self == other or self < other

    def __eq__(self, other):
        return self.line == other.line and self.column == other.column

    def __ne__(self, other):
        return not self == other

    def __gt__(self, other):
        return not self <= other

    def __ge__(self, other):
        return not self < other

    def __str__(self):
        return f"{self.line}.{self.column}"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash((self.line, self.column))



from __future__ import annotations

import os
import tkinter as tk
import typing

from biscuit.common.ui import Button, Frame, Label

if typing.TYPE_CHECKING:
    from .results import Results


class ExtensionGUI(Frame):
    """Extension item in the Extensions view.

    The Extension class represents an extension item in the Extensions view.
    """

    def __init__(
        self, master: Results, name: str, data: list[str], *args, **kwargs
    ) -> None:
        """Initialize the Extension class.

        Args:
            master (tk.Tk): The root window.
            name (str): The name of the extension.
            data (list): The extension data."""

        super().__init__(master, *args, **kwargs)
        self.master: Results = master
        self.config(**self.base.theme.views.sidebar.item)
        self.selected = False
        self.manager = self.base.extensions_manager

        self.data = data
        self.name = name
        self.filename = data[0]
        self.file = os.path.join(self.base.extensiondir, data[0])
        self.author = data[1]
        self.description = data[2][:50] + "..." if len(data[2]) > 30 else data[2]

        self.url = f"{self.manager.repo_url}extensions/{data[0]}"
        self.installed = os.path.isfile(self.file)

        self.bg = self.base.theme.views.sidebar.item.background
        self.hbg = self.base.theme.views.sidebar.item.highlightbackground

        self.container = Frame(self, padx=10, pady=10)
        self.container.pack(fill=tk.BOTH, expand=True)

        self.namelbl = Label(
            self.container,
            text=name[0].upper() + name[1:],
            font=("Segoi UI", 11, "bold"),
            anchor=tk.W,
            **self.base.theme.views.sidebar.item.content,
        )
        self.namelbl.pack(fill=tk.X)

        self.descriptionlbl = Label(
            self.container,
            text=self.description,
            font=("Segoi UI", 9),
            anchor=tk.W,
            **self.base.theme.views.sidebar.item.content,
        )
        self.descriptionlbl.config(fg="grey")
        self.descriptionlbl.pack(fill=tk.X, expand=True)

        self.subcontainer = Frame(self.container)
        self.subcontainer.pack(side=tk.BOTTOM, fill=tk.X, expand=True)

        self.authorlbl = Label(
            self.subcontainer,
            text=f"@{self.author}",
            font=("Segoi UI", 7, "bold"),
            anchor=tk.W,
            **self.base.theme.views.sidebar.item.content,
        )
        self.authorlbl.config(fg="grey")
        self.authorlbl.pack(side=tk.LEFT, fill=tk.X)

        self.install = Button(
            self.subcontainer,
            "Install",
            self.run_fetch_extension,
            padx=10,
        )
        self.install.config(font=("Segoi UI", 8), pady=2)
        self.install.pack(side=tk.RIGHT, fill=tk.X)
        if self.installed:
            self.install.config(text="Installed", bg=self.base.theme.biscuit_dark)
            self.install.set_command(self.remove_extension)

        self.bind("<Button-1>", self.set_selected)
        self.namelbl.bind("<Button-1>", self.set_selected)
        self.descriptionlbl.bind("<Button-1>", self.set_selected)
        self.authorlbl.bind("<Button-1>", self.set_selected)
        self.container.bind("<Button-1>", self.set_selected)
        self.subcontainer.bind("<Button-1>", self.set_selected)

        self.bind("<Enter>", self.hoverin)
        self.bind("<Leave>", self.hoveroff)
        self.hoveroff()

    def remove_extension(self, *_):
        self.manager.remove_extension(self)

    def run_fetch_extension(self, *_):
        self.manager.run_fetch_extension(self)

    def set_unavailable(self):
        self.install.config(text="Unavailable", bg=self.base.theme.biscuit_dark)

    def set_installed(self):
        self.install.config(text="Installed", bg=self.base.theme.biscuit_dark)
        self.install.set_command(self.remove_extension)

    def set_uninstalled(self):
        self.install.config(text="Install", bg=self.base.theme.biscuit)
        self.install.set_command(self.run_fetch_extension)

    def set_fetching(self):
        self.install.config(text="Fetching...", bg=self.base.theme.biscuit_dark)

    def set_selected(self, *_):
        self.master.set_selected(self)

    def select(self):
        self.selected = True
        self.config(bg=self.hbg)
        self.namelbl.config(bg=self.hbg)
        self.authorlbl.config(bg=self.hbg)
        self.descriptionlbl.config(bg=self.hbg)
        self.container.config(bg=self.hbg)
        self.subcontainer.config(bg=self.hbg)

    def deselect(self):
        self.selected = False
        self.config(bg=self.bg)
        self.namelbl.config(bg=self.bg)
        self.authorlbl.config(bg=self.bg)
        self.descriptionlbl.config(bg=self.bg)
        self.container.config(bg=self.bg)
        self.subcontainer.config(bg=self.bg)

    def hoverin(self, *_) -> None:
        if self.selected:
            return

        try:
            self.config(bg=self.hbg)
            self.namelbl.config(bg=self.hbg)
            self.authorlbl.config(bg=self.hbg)
            self.descriptionlbl.config(bg=self.hbg)
            self.container.config(bg=self.hbg)
            self.subcontainer.config(bg=self.hbg)
        except:
            pass

    def hoveroff(self, *_) -> None:
        if self.selected:
            return

        try:
            self.config(bg=self.bg)
            self.namelbl.config(bg=self.bg)
            self.authorlbl.config(bg=self.bg)
            self.descriptionlbl.config(bg=self.bg)
            self.container.config(bg=self.bg)
            self.subcontainer.config(bg=self.bg)
        except:
            pass



import os

import toml

from .theme.catppuccin_mocha import CatppuccinMocha
from .theme.gruvbox_dark import GruvboxDark

# from .theme import Dark, Light, Theme


class Config:
    """Loads and manages configurations for biscuit."""

    def __init__(self, master) -> None:
        self.base = master.base

        self.theme = GruvboxDark()
        self.font = ("Fira Code", 12)
        self.uifont = ("Fira Code", 10)

        self.auto_save_enabled = False
        self.auto_save_timer_ms = 10000

    def get_config_path(self, relative_path: str) -> str:
        """Get the absolute path to the resource

        Args:
            relative_path (str): path relative to the config directory"""

        return os.path.join(self.base.configdir, relative_path)

    def load_config(self) -> dict:
        with open(self.get_config_path("settings.toml"), "r") as settingsfile:
            config = toml.load(settingsfile)

        return config

    def load_data(self) -> None:
        # TODO testing
        self.theme = GruvboxDark()
        self.font = (self.config["font"], self.config["font_size"])



from __future__ import annotations

import typing

from .endpoint import Endpoint

if typing.TYPE_CHECKING:
    from biscuit.views import PanelView, SideBarView


class Views(Endpoint):
    """Views endpoint

    Side bar and panel views are managed by this endpoint.
    Further, new views can be added to the side bar and panel.

    Abstract `PanelView` and `SidebarView` are provided to create new views.
    """

    def __init__(self, *a) -> None:
        super().__init__(*a)
        self.sidebar = self.base.sidebar
        self.panel = self.base.sidebar

    def add_sidebar_view(self, view: PanelView):
        self.sidebar.add_view(view)

    def add_panel_view(self, view: SideBarView):
        self.panel.add_view(view)

    @property
    def explorer(self):
        return self.sidebar.explorer

    @property
    def search(self):
        return self.sidebar.search

    @property
    def source_control(self):
        return self.sidebar.source_control

    @property
    def logger(self):
        return self.panel.logger

    @property
    def terminal(self):
        return self.panel.terminal



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconLabelButton, WrappingLabel


class ExtensionsPlaceholder(Frame):
    """Placeholder view for the extensions view.

    The ExtensionsPlaceholder is displayed when the user has not yet fetched the extensions.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=15, pady=10, **self.base.theme.views.sidebar.item)
        self.columnconfigure(0, weight=1)

        WrappingLabel(
            self,
            text="Fetching extensions failed, check your internet connection.",
            font=("Segoe UI", 10),
            anchor=tk.W,
            **self.base.theme.views.sidebar.item.content
        ).grid(row=0, sticky=tk.EW)

        open_btn = IconLabelButton(
            self,
            text="Retry",
            icon=Icons.SYNC,
            callback=self.retry,
            pady=2,
            highlighted=True,
        )
        open_btn.grid(row=1, pady=5, sticky=tk.EW)

    def retry(self, *_) -> None:
        self.master.refresh()



from __future__ import annotations

import re
import tkinter as tk
import typing
from itertools import chain

if typing.TYPE_CHECKING:
    from biscuit.editor.text import Text
    from biscuit.language.data import Completion, Completions

from biscuit.common.ui import Toplevel

from .item import CompletionItem


class AutoComplete(Toplevel):
    """Floating window for autocomplete suggestions.

    In lsp mode, it receives completions from the language server.
    In regular mode, it generates completions from the current tab's words.

    NOTE: As of now, the window is limited to 10 items, not scrollable."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=2, pady=2, bg=self.base.theme.primary_background_highlight)
        self.withdraw()
        self.overrideredirect(True)
        self.wm_attributes("-topmost", True)

        self.lsp_mode = False
        self.latest_tab = None
        self.active = False
        self.selected = 0
        self.row = 0

        # 10 items are created and stored in menu_items, then moved to active_items
        # when needed. This is to avoid creating and destroying items on every update.
        self.menu_items: list[CompletionItem] = [
            CompletionItem(self) for _ in range(10)
        ]
        self.active_items: list[CompletionItem] = []

        self.row = 1

    def refresh_geometry(self, tab: Text) -> None:
        """Refresh the position of the autocomplete window.

        The window is placed right below the cursor, and it's width is
        adjusted to the longest item in the list.

        Args:
            tab (Text): The current tab."""

        self.update_idletasks()
        self.geometry("+{}+{}".format(*tab.cursor_screen_location()))

    def show(self, tab: Text) -> None:
        """Show the autocomplete window.

        Args:
            tab (Text): The current tab."""

        self.latest_tab = tab
        self.active = True
        self.update_idletasks()
        self.deiconify()

    def hide(self, *_):
        """Hide the autocomplete window."""

        self.active = False
        self.withdraw()
        self.reset()

    def reset(self):
        """Reset the autocomplete window for the next use."""

        self.reset_selection()

    def choose(self, tab: Text = None, this: CompletionItem = None):
        """Choose the currently selected item.

        Args:
            tab (Text, optional): The current tab. Defaults to latest_tab.
            this (CompletionItem, optional): The currently selected item. Defaults to selected item.
        """

        if not self.active_items:
            return

        tab = tab or self.latest_tab
        this = this or self.active_items[self.selected]

        # the language server provides the indices and word to replace with
        if self.lsp_mode:
            tab.replace(this.replace_start, this.replace_end, this.replace_text)
        else:
            tab.replace_current_word(this.replace_text)

        self.hide()
        return "break"

    def lsp_set_active_items(self, completions: list[Completion], term: str):
        """Set the active items with completions received from the language server.
        This is a helper function for lsp_update_completions.

        Args:
            completions (list[Completion]): The list of completions.
            term (str): The current word to complete."""

        # remove amount of items that are not needed
        while len(self.active_items) > len(completions):
            i = self.active_items.pop()
            i.grid_forget()
            self.menu_items.append(i)

        # add amount of items that are extra needed
        while len(self.active_items) < len(completions):
            i = self.menu_items.pop()
            i.grid(row=self.row, column=0, sticky="nsew")
            self.active_items.append(i)
            self.row += 1

        # now we have the same amount of items as completions
        # set the data for each item
        for i, completion in enumerate(completions):
            self.active_items[i].lsp_set_data(completion, term)

    def set_active_items(self, words: list[str], term: str):
        """Normal mode: Set the active items with words generated from the current tab.
        This is a helper function for update_completions.

        Args:
            words (list[str]): The list of words.
            term (str): The current word to complete."""

        while len(self.active_items) > len(words):
            i = self.active_items.pop()
            i.grid_forget()
            self.menu_items.append(i)

        while len(self.active_items) < len(words):
            i = self.menu_items.pop()
            i.grid(row=self.row, column=0, sticky="nsew")
            self.active_items.append(i)
            self.row += 1

        # now we have the same amount of items as words
        for i, word in enumerate(words):
            self.active_items[i].set_data(word, term)

    def clear(self) -> None:
        """Clear the active items and move them back to menu_items."""

        while self.active_items:
            i = self.active_items.pop()
            i.grid_forget()
            self.menu_items.append(i)

    def lsp_update_completions(self, tab: Text, completions: Completions) -> None:
        """Update the completions with the data received from the language server.

        Args:
            tab (Text): The current tab.
            completions (Completions): The completions received from the language server.
        """

        self.refresh_geometry(tab)

        term = tab.get_current_word()
        if completions:
            self.lsp_mode = True
            self.lsp_set_active_items(completions[:10], term)
            self.show(tab)
        else:
            self.hide()

    def update_completions(self, tab: Text):
        """Update the completions with words generated from the current tab.
        The words are filtered and prioritized based on exact match, starts with, and includes.

        Args:
            tab (Text): The current tab."""

        self.refresh_geometry(tab)

        term = tab.get_current_word()

        # term is too short
        if len(term) < 2:
            self.hide()
            return

        exact, starts, includes = [], [], []
        for word in tab.words:
            if word == term:
                exact.append(word)
            elif word.startswith(term):
                starts.append(word)
            elif term in word:
                includes.append(word)
        new = list(chain(exact, starts, includes))

        if new:
            self.lsp_mode = False
            self.set_active_items(new[:10], term)
            self.show(tab)
        else:
            self.hide()

    def select(self, delta):
        """Select the next or previous item.

        Args:
            delta (int): The direction to move the selection."""

        self.selected += delta
        if self.selected > len(self.active_items) - 1:
            self.selected = 0
        elif self.selected < 0:
            self.selected = len(self.active_items) - 1
        self.refresh_selected()

    def reset_selection(self):
        """Reset the selected item to the first one."""

        self.selected = 0
        self.refresh_selected()

    def refresh_selected(self):
        """Refresh the selected item."""

        for i in self.active_items:
            i.deselect()
        if self.selected < len(self.active_items):
            self.active_items[self.selected].select()

    def move_up(self, *_):
        """Move the selection up."""

        if self.active:
            self.select(-1)
            return "break"

    def move_down(self, *_):
        """Move the selection down."""

        if self.active:
            self.select(1)
            return "break"



from __future__ import annotations

import tkinter as tk
import typing
from tkinter import ttk

import tarts as lsp

from biscuit.common import ActionSet
from biscuit.common.ui import Frame

from .placeholder import OutlineTreePlaceholder
from .tree import Tree

if typing.TYPE_CHECKING:
    from biscuit.editor import Text


class OutlineTree(Frame):
    """Tree view that displays the outline of the active document.

    The OutlineTree view displays the outline of the active document.
    - The user can click on a symbol to navigate to it.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.actionset = ActionSet("Goto symbol in Editor", "@", [])

        # self.tree = ttk.Treeview(self, show="tree", columns=("kind", "pos"), displaycolumns='',
        #                          selectmode=tk.BROWSE, *args, **kwargs)
        self.tree = Tree(self)
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.tree.grid_remove()

        self.scrollbar = ttk.Scrollbar(
            self, orient=tk.VERTICAL, command=self.tree.yview, style="TreeScrollbar"
        )
        self.scrollbar.grid(row=0, column=1, sticky=tk.NS)
        self.scrollbar.grid_remove()

        self.tree.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.tree.yview)

        self.placeholder = OutlineTreePlaceholder(self)
        self.placeholder.grid(row=0, column=0, sticky=tk.NSEW)

        self.bind("<Double-Button-1>", self.goto_symbol)

    def clear(self) -> None:
        # self.tree.delete(*self.tree.get_children())
        self.tree.delete("1.0", tk.END)

    def update_symbols(
        self, tab: Text = None, response: list[lsp.DocumentSymbol] = None
    ) -> None:
        if not response:
            self.placeholder.show(tab)
            self.tree.grid_remove()
            self.scrollbar.grid_remove()
            return

        self.placeholder.grid_remove()
        self.tree.grid()
        self.scrollbar.grid()

        self.tree.config(state=tk.NORMAL)
        self.clear()
        self.tree.add_items(response)
        self.tree.config(state=tk.DISABLED)

    def add_items(self, parent: str, items: list[lsp.DocumentSymbol]) -> None:
        if not items:
            return

        for item in items:
            if item.kind == lsp.SymbolKind.MODULE:
                continue
            id = self.tree.insert(
                parent, "end", text=item.name, values=(item.range.start), open=True
            )
            self.add_items(id, item.children)

    def collapse_all(self, *_) -> None:
        for node in self.tree.get_children(""):
            self.tree.item(node, open=False)

    def goto_symbol(self, _: str) -> None:
        item = self.tree.item(self.tree.focus())
        if item:
            self.base.goto_location_in_active_editor(item.values[1])



import sv_ttk

from .theme import *


class VSCodeDark(Theme):
    name = "VSCode dark"

    border = "#2A2A2A"
    disabled = "#737373"

    primary_background = "#181818"
    primary_foreground = "#8B949E"
    primary_background_highlight = "#2C2D2D"
    primary_foreground_highlight = "#CCCCCC"

    secondary_background = "#1f1f1f"
    secondary_foreground = "#a9a9a9"
    secondary_background_highlight = "#323232"
    secondary_foreground_highlight = "#e0e0e0"

    # syntax
    Keyword = "#569cd6"
    Name = "#4EC9B0"
    Function = "#dcdcaa"
    String = "#ce9178"
    Number = "#b5cea8"
    Comment = {"foreground": "#7ca668", "font": {"slant": "italic"}}
    Punctuation = "#808080"

    def __init__(self, *args, **kwds) -> None:
        super().__init__(*args, **kwds)
        sv_ttk.use_dark_theme()

        self.editors.linenumbers.number.foreground = "#6e7681"
        self.editors.linenumbers.number.highlightforeground = "#cccccc"

        self.editors.diff.not_exist = "#424242"
        self.editors.diff.removed = "#452323"
        self.editors.diff.addition = "#203423"

        self.editors.currentword = "#474747"
        self.editors.found = "#203423"
        self.editors.selection = "#264f78"
        self.editors.currentword = "#27323c"
        self.editors.found = "#623315"
        self.editors.foundcurrent = "#9e86b7"
        self.editors.hovertag = "#202b35"

        self.views.panel.logs.time = "#6a9955"
        self.views.panel.logs.caller = "#569cd6"
        self.views.panel.logs.info = "#b5cea8"
        self.views.panel.logs.warning = "#ce9178"
        self.views.panel.logs.error = "#ce9178"

        self.editors.linenumbers.breakpoint.background = "#6e1b13"
        self.editors.linenumbers.breakpoint.highlightbackground = "#e51400"

        self.layout.statusbar.button_highlighted.update(
            bg=self.biscuit,
            fg="white",
            highlightbackground=self.biscuit_dark,
            highlightforeground="white",
        )



"""Layout

Contains the layout of the application. It is divided into 3 main parts:

Root
├── Menubar
├── <Container>
│  ├── Sidebar (Explorer, Search, Source Control, etc.)
│  └── Content
│     ├── Editors
│     └── Panel (Terminals, Logs, Problems, etc.)
└── StatusBar

The layout is designed to be modular and flexible. Each part can be easily replaced or modified.
Extensions can easily access and modify the layout.
"""

from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Frame
from biscuit.layout.statusbar import activitybar

from .content import *
from .grip import Grip
from .menubar import Menubar
from .secondary_sidebar import SecondarySideBar
from .sidebar import SideBar
from .statusbar import Statusbar

if typing.TYPE_CHECKING:
    from ..app import App


class Root(Frame):
    """Root of the application

    - Contains the Menubar, Sidebar, Content Pane, and StatusBar
    - Manages the resizing of the application
    """

    def __init__(self, base: App, *args, **kwargs) -> None:
        super().__init__(base, *args, **kwargs)
        self.config(bg=self.base.theme.border)

        container = Frame(self, bg=self.base.theme.border)

        self.menubar = Menubar(container)
        self.statusbar = Statusbar(container)

        subcontainer = Frame(container, bg=self.base.theme.border)
        self.content = Content(subcontainer)
        self.sidebar = SideBar(subcontainer, activitybar=self.statusbar.activitybar)
        self.secondary_sidebar = SecondarySideBar(
            subcontainer, activitybar=self.statusbar.secondary_activitybar
        )

        # Window Resizing Grips
        grip_w = Grip(self, "w", "left_side")
        grip_e = Grip(self, "e", "right_side")
        grip_n = Grip(container, "n", "top_side")
        grip_s = Grip(container, "s", "bottom_side")

        grip_w.pack(fill=tk.Y, side=tk.LEFT)
        container.pack(fill=tk.BOTH, expand=True, side=tk.LEFT)
        grip_e.pack(fill=tk.Y, side=tk.LEFT)

        grip_n.pack(fill=tk.X)
        self.menubar.pack()
        subcontainer.pack(fill=tk.BOTH, expand=True, pady=(1, 0))
        self.statusbar.pack()
        grip_s.pack(fill=tk.X)

        self.content.pack()
        self.pack(fill=tk.BOTH, expand=True)

    def toggle_sidebar(self) -> None:
        self.sidebar.toggle()



import tkinter as tk
import webbrowser

import mistune
from tkinterweb import HtmlFrame

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, Label, LinkLabel, Scrollbar
from biscuit.common.ui.buttons import IconLabelButton

from ..editor.editorbase import BaseEditor


class PRViewer(BaseEditor):
    """Pull Request Viewer

    Pull Request Viewer is a class that displays the details of a pull request in a GitHub repository.
    """

    def __init__(
        self,
        master,
        data: dict,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.filename = self.path = f"Pull request #{data['number']}"
        self.config(bg=self.base.theme.border)
        self.data = data

        label_cfg = {
            "bg": "#1C1C1C",
            "fg": self.base.theme.secondary_foreground,
            "font": (self.base.settings.uifont["family"], 12, "bold"),
        }

        container = Frame(self, width=150, padx=10, pady=10)

        # OPENED BY ------------------------------------------------------------

        Label(container, text="Opened by", **label_cfg).pack(
            side=tk.TOP, anchor=tk.W, pady=(20, 5)
        )
        LinkLabel(
            container,
            data["user"]["login"],
            lambda l=data["user"]["html_url"]: webbrowser.open(l),
        ).pack(side=tk.TOP, anchor=tk.W, pady=(0, 15)).config(bg="#1C1C1C")

        # ASSIGNEES ------------------------------------------------------------

        Label(container, text="Assignees", **label_cfg).pack(
            side=tk.TOP, anchor=tk.W, pady=5
        )
        for assignee in data["assignees"]:
            LinkLabel(
                container,
                assignee["login"],
                lambda l=assignee["html_url"]: webbrowser.open(l),
            ).pack(side=tk.TOP, anchor=tk.W).config(bg="#1C1C1C")

        # LABELS --------------------------------------------------------------

        Label(container, text="Labels", **label_cfg).pack(
            side=tk.TOP, anchor=tk.W, pady=(20, 5)
        )

        label_url = (
            "/".join(data["html_url"].split("/")[:-1]) + "?q=is%3Aissue+label%3A{}"
        )

        subcontainer = Frame(container)
        subcontainer.pack(side=tk.TOP, anchor=tk.W, pady=5)
        inline = 0
        for label in data["labels"]:
            if inline > 2:
                subcontainer = Frame(container)
                subcontainer.pack(side=tk.TOP, anchor=tk.W, pady=5)
                inline = 0

            # color = "#" + label["color"]
            IconLabelButton(
                subcontainer,
                label["name"],
                icon=Icons.TAG,
                callback=lambda l=label_url.format(label["name"]): webbrowser.open(l),
                # bg=color,
                # hbg=color,
            ).pack(side=tk.LEFT, padx=5)
            inline += 1

        # BODY --------------------------------------------------------------

        self.body = HtmlFrame(self, messages_enabled=False, vertical_scrollbar=False)
        self.scrollbar = Scrollbar(
            self,
            orient=tk.VERTICAL,
            command=self.body.yview,
            style="EditorScrollbar",
        )
        self.body.html.config(yscrollcommand=self.scrollbar.set)
        self.body.html.shrink(True)

        def load_new_page(url: str):
            webbrowser.open(url)

        self.body.on_link_click(load_new_page)

        self.body.load_html(
            f"<h1>{data['title']} <a href={data['html_url']}>#{data['number']}</a></h1>"
            + f"<span class='state-label'>{data['state']}</span>"
            + f"<a href={data['user']['html_url']}>{data['user']['login']}</a>"
            + f" wants to merge <code>{data['head']['label']}</code> into <code>{data['base']['label']}</code>"
            + "<br><hr>"
            + mistune.html(data["body"])
        )

        # TODO: load comments

        t = self.base.theme
        css = f"""
            img {{
                max-width: 100%;
                height: auto;
            }}

            
            hr {{
                border: 0;
                border-top: 1px solid {t.border};
                max-width: 100%;
            }}

            CODE, PRE {{
                font-family: {self.base.settings.font['family']};
                font-size: {self.base.settings.font['size']}pt;
                background-color: {t.border};
                padding: 2px;
            }}
            BODY {{
                background-color: {t.secondary_background};
                color: {t.secondary_foreground};
                padding: 10px;
                padding-left: 20px;
                padding-right: 20px;
            }}
            
            span.state-label {{
                display: inline-block;
                padding: 0.25em 0.5em;
                margin-right: 10px;
                font-size: 1.1em;
                font-weight: bold;
                color: white;
                background-color: {t.biscuit};
            }}

            li{{
                margin-left:1px;
            }}
            :link    {{ color: {t.biscuit}; }}
            :visited {{ color: {t.biscuit_dark}; }}
            INPUT, TEXTAREA, SELECT, BUTTON {{ 
                background-color: {t.secondary_background};
                color: {t.secondary_foreground_highlight};
            }}
            INPUT[type="submit"],INPUT[type="button"], INPUT[type="reset"], BUTTON {{
                background-color: {t.primary_background};
                color: {t.primary_foreground};
                color: tcl(::tkhtml::if_disabled {t.primary_background}{t.primary_foreground_highlight});
            }}
            """
        self.body.add_css(css)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(1, weight=2)
        self.columnconfigure(3, weight=2)
        self.body.grid(row=0, column=1, sticky=tk.NSEW)
        self.scrollbar.grid(row=0, column=2, sticky=tk.NS, padx=(0, 1))
        container.grid(row=0, column=3, sticky=tk.NSEW)



import os
import tkinter as tk
from threading import Thread

if os.name == "nt":
    from winpty import PtyProcess as PTY
else:
    from ptyprocess import PtyProcessUnicode as PTY

from biscuit.common.ui import Scrollbar

from ..panelview import PanelView
from .ai import AI
from .ansi import replace_newline, strip_ansi_escape_sequences
from .text import TerminalText


class TerminalBase(PanelView):
    """Base class for all terminals

    Attributes:
        name (str): The name of the terminal.
        icon (str): The icon of the terminal.
        shell (str): The shell to use.
        p (PTY): The PTY process.

    Methods:
        __init__: Initializes the terminal.
        start_service: Starts the terminal service.
        destroy: Destroys the terminal.
        run_command: Runs a command in the terminal.
        enter: Handles the enter key press.
        write_loop: Writes the terminal output.
        insert: Inserts text into the terminal.
        newline: Inserts a new line.
        clear: Clears the terminal.
        ctrl_key: Handles the ctrl key press.
        __str__: Returns the name of the terminal.
    """

    name: str
    icon: str
    shell: str
    p: PTY

    def __init__(self, master, cwd=".", *args, **kwargs) -> None:
        """Initialize the terminal

        Args:
            master: The parent widget
            cwd: The current working directory"""

        super().__init__(master, *args, **kwargs)

        # MOVED to terminal manager
        # self.__actions__ = (("add",), ("trash", self.destroy))

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.alive = False
        self.cwd = cwd
        self.last_command = ""
        self.last_command_index = ""
        # self.prediction = ""

        self.text = TerminalText(
            self, relief=tk.FLAT, padx=10, pady=10, font=("Consolas", 11)
        )
        self.text.grid(row=0, column=0, sticky=tk.NSEW)
        self.text.bind("<Return>", self.enter)

        self.terminal_scrollbar = Scrollbar(self, style="EditorScrollbar")
        self.terminal_scrollbar.grid(row=0, column=1, sticky="NSW")

        self.text.config(yscrollcommand=self.terminal_scrollbar.set)
        self.terminal_scrollbar.config(command=self.text.yview, orient=tk.VERTICAL)

        self.text.tag_config("prompt", foreground=self.base.theme.biscuit_dark)
        self.text.tag_config("command", foreground=self.base.theme.biscuit)
        self.text.tag_config("ghost", foreground=self.base.theme.border)

        self.ai = AI(self)

        self.bind("<Destroy>", self.destroy)
        # self.bind("<Tab>", self.tab_key)

    def start_service(self, *_) -> None:
        self.alive = True
        self.last_command = None

        self.p = PTY.spawn([self.shell], cwd=self.cwd)
        Thread(target=self.write_loop, daemon=True).start()

    def destroy(self, *_) -> None:
        self.alive = False

    def run_command(self, command: str) -> None:
        self.text.insert("end", command, "command")
        self.enter()

    def enter(self, *_) -> None:
        self.last_command_index = self.text.index("input")

        command = self.text.get("input", "end")

        if command.startswith("#"):
            self.text.delete("input", "end")
            self.ai.get_response(command)
            return "break"

        self.last_command = command
        self.text.register_history(command)
        if command.strip():
            self.text.delete("input", "end")
            self.text.tag_add("prompt", "input linestart", "input")

        self.p.write(command + "\r\n")
        return "break"

    def write_loop(self) -> None:
        while self.alive:
            if buf := self.p.read():
                p = buf.find("\x1b]0;")

                if p != -1:
                    buf = buf[:p]
                buf = [
                    strip_ansi_escape_sequences(i)
                    for i in replace_newline(buf).splitlines()
                ]
                buf = "\n".join(buf)

                self.insert(buf)
                if "is not recognized as an internal or external command" in buf:
                    self.error()

    def error(self):
        self.ai.get_response(self.last_command)

    def insert(self, output: str, tag="") -> None:
        self.text.insert(tk.END, output, tag)
        # self.terminal.tag_add("prompt", "insert linestart", "insert")
        self.text.see(tk.END)
        self.text.mark_set("input", "insert")

    def newline(self):
        self.insert("\n")

    def clear(self) -> None:
        self.text.clear()

    # def ghost_insert(self, output: str) -> None:
    #     self.text.insert(tk.END, output, "ghost")
    #     self.text.see(tk.END)
    #     self.prediction = output

    # def tab_key(self, *_) -> None:
    #     if t := self.text.get("input", "insert"):
    #         if t.strip() in self.prediction.strip():
    #             self.text.delete("input", "end")
    #             self.text.insert("end", self.prediction)
    #     else:
    #         self.text.insert("end", "    ")

    def ctrl_key(self, key: str) -> None:
        if key == "c":
            self.run_command("\x03")

    def __str__(self) -> str:
        return self.name



"""Utility functions for the Python debugger."""

import ctypes
import os
import sys
import types
import typing


def get_callstack(
    frame: types.FrameType,
) -> typing.Generator[tuple[str, str, int], None, None]:
    """Returns the call stack as a list of tuples (function name, filename, line number)."""

    while frame:
        yield (
            frame.f_code.co_name,
            os.path.abspath(frame.f_code.co_filename),
            frame.f_lineno,
        )
        frame = frame.f_back


def get_variables(
    frame: types.FrameType,
) -> typing.Generator[tuple[str, dict, bool | None], None, None]:
    """Returns section name, variable mapping, and whether tree nodes should be expanded."""

    code_info = {
        "name": frame.f_code.co_name,
        "filename": frame.f_code.co_filename,
        "firstlineno": frame.f_code.co_firstlineno,
        "argcount": frame.f_code.co_argcount,
        "posonlyargcount": frame.f_code.co_posonlyargcount,
        "kwonlyargcount": frame.f_code.co_kwonlyargcount,
        "nlocals": frame.f_code.co_nlocals,
        "stacksize": frame.f_code.co_stacksize,
        "flags": frame.f_code.co_flags,
        "code": repr(frame.f_code.co_code),  # Use repr for bytes
        "consts": frame.f_code.co_consts,
        "names": frame.f_code.co_names,
        "varnames": frame.f_code.co_varnames,
        "freevars": frame.f_code.co_freevars,
        "cellvars": frame.f_code.co_cellvars,
    }

    if sys.version_info >= (3, 10):
        code_info["linetable"] = repr(frame.f_code.co_linetable)
        code_info["lines"] = list(frame.f_code.co_lines())
    else:
        code_info["lnotab"] = repr(frame.f_code.co_lnotab)

    if sys.version_info >= (3, 11):
        code_info["exceptiontable"] = repr(frame.f_code.co_exceptiontable)
        code_info["qualname"] = frame.f_code.co_qualname
        code_info["positions"] = list(frame.f_code.co_positions())

    yield "Code Info", code_info
    yield "Globals", frame.f_globals
    yield "Locals", frame.f_locals, True


# CREDITS: https://pydev.blogspot.com/2014/02/changing-locals-of-frame-frameflocals.html
def save_locals(frame: types.FrameType) -> dict:
    ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_int(0))
    return frame.f_locals
